/*    Audio-CD  WAV- */

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/cdrom.h>
#include <linux/types.h>
#include <sys/stat.h>

#define CD_DEVICE "/dev/cdrom"
#define WAV_HDR_LEN 44

/*    WAV- */
typedef struct {
    __u8 riff[4];
    __u32 size;
    __u8 wave[8];
    __u32 size1;
    __u16 format_tag;
    __u16 channels;
    __u32 sample_per_sec;
    __u32 byte_per_sec;
    __u16 block_align;
    __u16 bit_per_sample;
    __u8 data[4];
    __u32 size2;
} __attribute__ ((packed)) wav_header_t;

wav_header_t w_hdr;

int main()
{
    int fd, out, n;
    unsigned int i, start_lba, end_lba;

    __u8 buff[CD_FRAMESIZE_RAW];

    struct stat s;
    struct cdrom_tochdr hdr;
    struct cdrom_tocentry toc;
    struct cdrom_read_audio cda;

    fd = open(CD_DEVICE, O_RDONLY|O_NONBLOCK);
    if(fd < 0) {
	perror("open");
	return -1;
    }

/*   -.    Audio-CD */
    if(ioctl(fd, CDROM_DISC_STATUS) != CDS_AUDIO) {
	printf("I need Audio_CD!\n");
	return 0;
    }

    memset((void *)&hdr, 0, sizeof(struct cdrom_tochdr));
    if(ioctl(fd, CDROMREADTOCHDR, &hdr) < 0) {
	perror("ioctl");
	return(errno);
    }

    printf("First: %d\t", hdr.cdth_trk0);
    printf("Last: %d\n", hdr.cdth_trk1);

#define FIRST hdr.cdth_trk0
#define LAST hdr.cdth_trk1

    printf("Enter track number: ");
    scanf("%d", &n);

    if((n < 1) || (n > LAST)) {
	printf("Wrong track number\n");
	return -1;
    }

/*     LBA  */
    toc.cdte_format = CDROM_LBA;
    toc.cdte_track = n;

    if(ioctl(fd, CDROMREADTOCENTRY, &toc) < 0) {
        perror("ioctl");
        return -1;
    }

    start_lba = toc.cdte_addr.lba;

/*    */
    if(n == LAST) toc.cdte_track = CDROM_LEADOUT;
    else toc.cdte_track = n + 1;

    if(ioctl(fd, CDROMREADTOCENTRY, &toc) < 0) {
        perror("ioctl");
        return -1;
    }

    end_lba = toc.cdte_addr.lba;

    out = open("./track.wav", O_CREAT|O_RDWR|O_TRUNC, 0600);
    if(out < 0) { perror("open"); return -1; }

    memset(&w_hdr, 0, sizeof(wav_header_t));
    write(out, (void *)&w_hdr, WAV_HDR_LEN);

    cda.addr_format = CDROM_LBA;
    cda.nframes = 1;
    cda.buf = buff;

    printf("Track size - %d sectors\n", end_lba - start_lba);

    for(i = start_lba; i < end_lba; i++) {

        memset(buff, 0, sizeof(buff));

	cda.addr.lba = i;
	fprintf(stdout,"\rLBA: %u", i - start_lba + 1);
	fflush(stdout);

	if(ioctl(fd, CDROMREADAUDIO, &cda) < 0) {
	    perror("ioctl");
	    return -1;
	}

	if(write(out, (__u8 *)buff, CD_FRAMESIZE_RAW) < 0) {
	    perror("write");
	    return -1;
	}
    }

    printf("\n");

    memset(&s, 0, sizeof(struct stat));
    if(fstat(out, &s) < 0) {
	perror("fstat");
	exit(-1);
    }

/* Make WAV header */
    memset(&w_hdr, 0, sizeof(wav_header_t));
    memcpy(w_hdr.riff, "RIFF", 4);
    w_hdr.size = s.st_size - 8;
    memcpy(w_hdr.wave, "WAVEfmt ", 8); //   - 
    w_hdr.size1 = 16;
    w_hdr.format_tag = 1;
    w_hdr.channels = 2;
    w_hdr.sample_per_sec = 44100;
    w_hdr.byte_per_sec = 176400;
    w_hdr.block_align = 4;
    w_hdr.bit_per_sample = 16;
    memcpy(w_hdr.data, "data", 4); 
    w_hdr.size2 = s.st_size - WAV_HDR_LEN;

    lseek(out, 0, 0);
    write(out, (void *)&w_hdr, WAV_HDR_LEN);

    printf("OK\n");

    close(fd);
    close(out);

    return 0;
}
