/*  Q-  */

#include <stdio.h>
#include <fcntl.h>
#include <linux/types.h>
#include <linux/cdrom.h>
#include <linux/byteorder/swab.h>

#define CDROM "/dev/cdrom"

int fd; //  
#define MSF2LBA(Min, Sec, Frame) (((Min * 60 + Sec) * 75 + Frame) - 150)

/*        */
typedef struct {
    __u8 dfc;		// Data Format Code
    __u8 ctrl	:4;	// Control
    __u8 adr	:4;	// ADR
    __u8 tno;		// Track number
    __u8 ino;		// Index number
    __u8 a_addr[4];	// Absolute CD-ROM Address
    __u8 r_addr[4];	// Track Relative CD-ROM Address
} cur_pos_t;


void test_unit_ready()
{
    int i;
    __u8 test_unit_cmd[CDROM_PACKET_SIZE];

    memset(test_unit_cmd, 0, 12);
    test_unit_cmd[0] = GPCMD_TEST_UNIT_READY;

    for(i = 0; i < 20; i++) {
	if(send_cmd(fd, test_unit_cmd, CGC_DATA_NONE, NULL, 0) == 0) return;
	sleep(1);
    }

    printf("\n  .\n");
    exit(-1);
}


int seek(__u32 lba)
{
    __u8 cmd[12];

    memset(cmd, 0, 12);

    cmd[0] = GPCMD_SEEK; //   SEEK (0x2B)
    lba = __swab32(lba);
    memcpy((void *)(cmd + 2), (void *)&lba, 4);

    test_unit_ready();

    if(send_cmd(fd, cmd, CGC_DATA_NONE, NULL, 0) < 0) return -1;

    return 0;
}


/*     */
int read_subch()
{
    __u8 cmd[CDROM_PACKET_SIZE];
    __u8 data_buff[16]; //    Q-
    cur_pos_t *cur_pos;
    __u32 lba = 0;

    memset(data_buff, 0, 16);
    memset(cmd, 0, sizeof(cmd));
    cmd[0] = GPCMD_READ_SUBCHANNEL;
    cmd[1] = 2; // .   MSF 
    cmd[2] = 0x40; //  SUBQ  -  Q- 
    cmd[3] = 1; //      .
    cmd[8] = 16;

    test_unit_ready();

    if(send_cmd(fd, cmd, CGC_DATA_READ, data_buff, 16) < 0) return -1;

    cur_pos = (void *)(data_buff + 4);

#define MIN cur_pos->a_addr[1]
#define SEC cur_pos->a_addr[2]
#define FRAME cur_pos->a_addr[3]

    printf("%d\t", cur_pos->ctrl);
    printf("%d\t", cur_pos->adr);
    printf("%X\t", cur_pos->tno);
    printf("%d\t", MIN);
    printf("%d\t", SEC);
    printf("%d\t", FRAME);

    lba = MSF2LBA(MIN, SEC, FRAME);
    printf("%u\n", lba);

    return;
}



int main(int argc, char **argv)
{
    int i;
    __u32 pos = 6655; //      

    fd = open(CDROM, O_RDONLY|O_NONBLOCK);
    if(fd < 0) { perror("open"); return -1; }

    printf("NUM\tSECT\tCTRL\tADR\tTNO\tMIN\tSEC\tFRAME\tLBA\n");

    for(i = 0; i < 30; i++) {
	printf("%d\t%u\t", i, pos);
	seek(pos);
	sleep(2); // ,        
	read_subch();
	pos += 1;
    }

    return 0;
}
