/*  RAW TOC CD- */

#include <stdio.h>
#include <fcntl.h>
#include <linux/types.h>
#include <linux/cdrom.h>
#include <linux/byteorder/swab.h>

#define CDROM "/dev/cdrom"
#define MSF2LBA(Min, Sec, Frame) (((Min * 60 + Sec) * 75 + Frame) - 150)

int fd; //  

struct toc {
    __u8 snum;		// Session Number
    __u8 ctrl	:4;	// Control
    __u8 adr	:4;	// ADR
    __u8 tno;		// TNO (always 0)
    __u8 point;		// POINT
    __u8 min;		// AMIN
    __u8 sec;		// ASEC
    __u8 frame;		// AFRAME
    __u8 zero;		// 0
    __u8 pmin;		// PMIN
    __u8 psec;		// PSEC
    __u8 pframe;	// PFRAME
};

/* Read TOC command */
int read_raw_toc()
{
    int i;
    __u8 cmd[12];
    __u8 *data_buff; //      TOC
    __u16 toc_data_len = 0; //   TOC
    int toc_entries = 0; //   TOC
    int last; //   /
    __u32 lba;
    struct toc *t;

    memset(cmd, 0, 12);
    data_buff = (__u8 *)malloc(0xFFFF);
    memset(data_buff, 0, 0xFFFF);

    cmd[0] = GPCMD_READ_TOC_PMA_ATIP;
    cmd[2] = 2; // Format Field = 10b
    cmd[7] = 0xFF;
    cmd[8] = 0xFF;

    if(send_cmd(fd, cmd, CGC_DATA_READ, data_buff, 0xFFFF) < 0) return -1;

    *((__u8 *)&toc_data_len) = data_buff[1];
    *((__u8 *)&toc_data_len + 1) = data_buff[0];
    printf("TOC data length: %d\n", toc_data_len);

    toc_entries = (toc_data_len - 2)/11;
    printf("TOC entries: %d\n", toc_entries);

    printf("First: %d\t", data_buff[2]);
    last = data_buff[3];
    printf("Last: %d\n", last);

    t = (struct toc *)malloc(toc_data_len);
    memset((void *)t, 0, toc_data_len);
    memcpy((void *)t, data_buff + 4, toc_data_len);

    free(data_buff);

#define SNUM(i) (t + i)->snum
#define ADR(i) (t + i)->adr
#define CTRL(i) (t + i)->ctrl
#define PMIN(i) (t + i)->pmin
#define PSEC(i) (t + i)->psec
#define PFRAME(i) (t + i)->pframe
#define MIN(i) (t + i)->min
#define SEC(i) (t + i)->sec
#define FRAME(i) (t + i)->frame
#define POINT(i) (t + i)->point

    printf("\n    - %d\n", last);
    printf("Entry\tSession\tADR\tCTRL\tPoint\tMin\tSec\tFrame\tPMin\tPsec\tPFrame\tLBA\n");

    for(i = 0; i < toc_entries; i++) {
	printf("%d\t", i);
	printf("%d\t", SNUM(i));
	printf("%d\t", ADR(i));
	printf("%d\t", CTRL(i));
	printf("%X\t", POINT(i));
	printf("%d\t", MIN(i));
	printf("%d\t", SEC(i));
	printf("%d\t", FRAME(i));
	printf("%d\t", PMIN(i));
	printf("%d\t", PSEC(i));
	printf("%d\t", PFRAME(i));

	if((POINT(i) == 0xC0) || (POINT(i) == 0xC1)) {
	    printf("\n");
	    continue;
	}

	if(POINT(i) == 0xB0) lba = MSF2LBA(MIN(i), SEC(i), FRAME(i)) + 150;
	else lba = MSF2LBA(PMIN(i), PSEC(i), PFRAME(i));

	printf("%u\n", lba);
    }

    free(t);
    return 0;
}



int main()
{
    fd = open(CDROM, O_RDONLY|O_NONBLOCK);
    if(fd < 0) { perror("open"); return -1; }

    read_raw_toc();

    return 0;
}
