/*  TOC  c    */

#include <stdio.h>
#include <fcntl.h>
#include <linux/types.h>
#include <linux/cdrom.h>
#include <linux/byteorder/swab.h>

#define CDROM "/dev/cdrom"

int fd; //  
int max_track_num = 0; //    CDROM

struct toc {
    __u8 res;
    __u8 adr_cntl;
    __u8 trk_num;
    __u8 res1;
    __u32 lba;
    struct toc *next;
};

struct toc_entry {
    __u32 start_lba;
    __u32 end_lba;
};


/* Read TOC command */
struct toc * read_toc()
{
    static int i = 1;
    __u8 cmd[12];
    __u8 data[12]; // 12  - 4   + 8    
    struct toc *t;

    if(i > 0xAA) return NULL;

    t = (struct toc *)malloc(sizeof(struct toc));

    memset(cmd, 0, 12);
    memset(data, 0, 12);

    cmd[0] = GPCMD_READ_TOC_PMA_ATIP;
    cmd[6] = i;
    cmd[8] = 12;

    send_cmd(fd, cmd, CGC_DATA_READ, data, 12);

    if(i == 1) {
        printf("First: %d\t", data[2]);
        printf("Last: %d\n", data[3]);
	max_track_num = data[3];
    }

    memcpy((void *)t, (data + 4), 8);
    t->lba = __swab32(t->lba);

    i += 1;
    if(i == (max_track_num + 1)) i = 0xAA;

    t->next = read_toc();

    return t;
}


void view_toc(struct toc *t)
{
    if(t == NULL) return;

    if(t->trk_num == 0xAA) printf("lead out:\t");
    else printf("track: %d\t", t->trk_num);
    printf("lba: %u\n", t->lba);

    view_toc(t->next);
}


struct toc_entry * get_toc_entry(int trk_num, struct toc *t)
{
    struct toc_entry *t_entry;
    int i = 1;

    for(;i < trk_num; i++) t = t->next;

    t_entry = (struct toc_entry *)malloc(sizeof(struct toc_entry));
    t_entry->start_lba = t->lba;
    t_entry->end_lba = t->next->lba;

    return t_entry;
}


int main()
{
    int trk_num = 0;
    struct toc *t = NULL;
    struct toc_entry *t_entry;

    fd = open(CDROM, O_RDONLY|O_NONBLOCK);
    if(fd < 0) { perror("open"); return -1; }

    t = read_toc();
    view_toc(t);

    printf("Enter track number: ");
    scanf("%d", &trk_num);

    if((trk_num < 1) || (trk_num > max_track_num)) {
	printf("Wrong track number\n");
	return -1;
    }

    t_entry = get_toc_entry(trk_num, t);
    printf("Start LBA - %u\n", t_entry->start_lba);
    printf("End LBA - %u\n", t_entry->end_lba);

    return 0;
}
