#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>

#define SG_DEV "/dev/sg1"
int sg_fd; //  

void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);
    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    test_unit_ready(sg_fd);
    send_cmd(sg_fd, start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 20);

    return;
}


/* /  */
int prevent_allow(__u8 prevent_flag)
{
    __u8 prevent_allow_cmd[6];

    memset(prevent_allow_cmd, 0, 6);
    prevent_allow_cmd[0] = 0x1E;
    prevent_allow_cmd[4] = prevent_flag;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, prevent_allow_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) < 0) return -1;

    return 0;
}


int blank(__u8 blank_type)
{
    __u8 blank_cmd[12];

    memset(blank_cmd, 0, 12);
    blank_cmd[0] = 0xA1; // BLANK
    blank_cmd[1] = blank_type;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, blank_cmd, 12, SG_DXFER_NONE, NULL, 0, 9600) < 0) return -1;

    return 0;
}


void usage()
{
    fprintf(stderr, "\n\tUsage: blank [mode]\n");
    fprintf(stderr, "\tMode: 0 - full blank, 1 - minimally blank\n\n");
    exit(0);
}


int main(int argc, char **argv)
{
    char mode = 0;

#define FULL_BLANK 0x00
#define MIN_BLANK 0x01

    if(argc != 2) usage();

    mode = atoi(argv[1]);
    if((mode < 0) || (mode > 1)) usage();

    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return -1;
    }

    prevent_allow(1); //  

/*    */
    printf("Blanking disk..\n");
    if(blank(mode) < 0) printf("-ERR: blank disk\n");
    else printf("OK\n");

    prevent_allow(0);

    eject_cd();
    close(sg_fd);

    return 0;
}
