#include <fcntl.h>
#include <syslog.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/types.h>

int send_cmd(int sg_fd, __u8 *cmd, __u8 cmdlen, int direction, __u8 *data, \
	    __u32 datalen, unsigned int timeout)
{
    sg_io_hdr_t io_hdr;
    __u8 sense_buffer[32];

#define SK (sense_buffer[2] & 0x0F)
#define ASC sense_buffer[12]
#define ASCQ sense_buffer[13]

    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cmdlen;
    io_hdr.mx_sb_len = sizeof(sense_buffer);
    io_hdr.dxfer_direction = direction;
    io_hdr.dxfer_len = datalen; //  
    io_hdr.dxferp = data; //    
    io_hdr.cmdp = cmd;
    io_hdr.sbp = sense_buffer;
    io_hdr.timeout = timeout * 1000; /* 20000 millisecs == 20 seconds */

rep:
    if(ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
    	perror("SG_IO ioctl");
    	return -1;
    }

    if((io_hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK) {
        if(io_hdr.sb_len_wr > 0) {

/*  SK/ASC/ASCQ == 02/04/08 (Not Ready. Long Write in Progress),
 *     WRITE_10
 */
	    if((SK == NOT_READY) && (ASC == 0x04) && (ASCQ == 0x08)) goto rep;
            syslog(LOG_INFO, "Command: 0x%02X", io_hdr.cmdp[0]);
            syslog(LOG_INFO, "Sense data (SK/ASC/ASCQ): 0x%02X/0x%02X/0x%02X", SK, ASC, ASCQ);
        }

        if(io_hdr.masked_status)
            syslog(LOG_INFO, "SCSI status = 0x%x\n", io_hdr.status);
        if(io_hdr.host_status)
	    syslog(LOG_INFO, "Host status = 0x%x\n", io_hdr.host_status);
        if(io_hdr.driver_status)
	    syslog(LOG_INFO, "Driver status = 0x%x\n", io_hdr.driver_status);

	return -1;
    }

    return 0;
}


void test_unit_ready(int sg_fd)
{
    int i;
    __u8 test_unit_cmd[6];

    /* Prepare TEST UNIT command */
    memset(test_unit_cmd, 0, 6);

    for(i = 0; i < 20; i++) {
	if(send_cmd(sg_fd, test_unit_cmd, 6, SG_DXFER_NONE, NULL, 0, 200) == 0) return;
	sleep(1);
    }

    printf("\nUnit not ready.\n");
    exit(-1);
}
