#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include <linux/cdrom.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include "read_disk_info.h"

#define SG_DEV "/dev/sg1"
int sg_fd;

disk_info_block_t *DIB;

int read_track_info(int trk_num)
{
    __u8 read_track_info_cmd[10];
    __u16 blk_size = 0;
    trk_info_t *trk_info;

    trk_info = (trk_info_t *)malloc(TRK_INFO_LEN);
    if(trk_info == NULL) { perror("malloc"); exit(errno); }

    memset(read_track_info_cmd, 0, 10);
    read_track_info_cmd[0] = 0x52;
    read_track_info_cmd[1] = 1;
    read_track_info_cmd[5] = trk_num;
    read_track_info_cmd[7] = TRK_INFO_LEN;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, read_track_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)trk_info, TRK_INFO_LEN, 200) < 0) return -1;

    *((__u8 *)&blk_size) = *((__u8 *)&trk_info->trk_info_len + 1);
    *((__u8 *)&blk_size + 1) = *((__u8 *)&trk_info->trk_info_len);

    printf("Track info size: %u\n", blk_size + 2);
    printf("Track number: %d\n", (trk_info->trk_num_ms << 8) | trk_info->trk_num_ls);
    printf("Start address of track: %u\n", __swab32(trk_info->start_addr));
    printf("Next writable address: %u\n", __swab32(trk_info->nwa));
    printf("Track size: %u\n", __swab32(trk_info->trk_size));
    printf("Free blocks: %u\n", __swab32(trk_info->free_blk));
    printf("RT Blank PAC/INC FP: %d %d %d %d\n\n", trk_info->RT, trk_info->blank, trk_info->PAC_INC, trk_info->FP);

    return 0;
}


int read_disk_info()
{
    int i = 0;
    __u8 read_disk_info_cmd[10];
    __u16 disk_info_len = 0;
    __u16 trk_num = 0;

    DIB = (disk_info_block_t *)malloc(DIB_LEN);
    if(DIB == NULL) { perror("malloc"); exit(errno); }

    memset(DIB, 0, DIB_LEN);
    memset(read_disk_info_cmd, 0, 10);
    read_disk_info_cmd[0] = 0x51;
    read_disk_info_cmd[8] = DIB_LEN;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, read_disk_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)DIB, DIB_LEN, 200) < 0) return -1;

    *((__u8 *)&disk_info_len) = *((__u8 *)&DIB->disk_info_len + 1);
    *((__u8 *)&disk_info_len + 1) = *((__u8 *)&DIB->disk_info_len);

    trk_num = (DIB->last_trk_ms << 8) | DIB->last_trk_ls;

    printf("Data info length: %d\n", disk_info_len);
    printf("Erasable: %d\n", DIB->erasable);
    printf("Disk status: 0x%.2x\n", DIB->disk_status);
    printf("Number of first track: %d\n", DIB->first_track_num);
    printf("Number of Session: %d\n", (DIB->sess_num_msb << 8) | DIB->sess_num_lsb);
    printf("Status of last Session: 0x%.2x\n", DIB->ls_status);
    printf("First track in last Session: %d\n", (DIB->first_trk_ms << 8) | DIB->first_trk_ls);
    printf("Last track in last Session: %d\n", trk_num);
    printf("BG format status: %d\n\n", DIB->BG_format_status);

    if(DIB->disk_status == 0) printf("Empty disk\n");
    else for(i = 1; i <= trk_num; i++) read_track_info(i);

    return 0;
}


int main(int argc, char **argv)
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return -1;
    }

    read_disk_info();

    close(sg_fd);
    return 0;
}
