#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/cdrom.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>

#define SG_DEV "/dev/sg0"

int sg_fd;

typedef struct {
    __u8 part_version	:4;
    __u8 book_type	:4;
    __u8 max_rate	:4;
    __u8 disk_size	:4;
    __u8 layer_type	:4;
    __u8 track_path	:1;
    __u8 num_of_layers	:2;
    __u8 rez1		:1;
    __u8 track_dns	:4;
    __u8 linear_dns	:4;
    __u32 start_psn_da;
    __u32 end_psn_da;
    __u32 end_psn_L0;
    __u8 rez2		:7;
    __u8 BCA		:1;
} __attribute__ ((packed)) layer_descr_t;


int send_cmd(__u8 *cmd, __u8 cmdlen, int direction, __u8 *data, \
	    __u32 datalen, unsigned int timeout)
{
    int k = 0;
    sg_io_hdr_t io_hdr;
    __u8 sense_buffer[32];

    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cmdlen;
    io_hdr.mx_sb_len = sizeof(sense_buffer);
    io_hdr.dxfer_direction = direction; //   
    io_hdr.dxfer_len = datalen; //  
    io_hdr.dxferp = data; //    
    io_hdr.cmdp = cmd;
    io_hdr.sbp = sense_buffer;
    io_hdr.timeout = timeout * 1000; /* 20000 millisecs == 20 seconds */

    if(ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
    	perror("SG_IO ioctl");
    	return -1;
    }

    /* now for the error processing */
    if((io_hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK) return -1;

    return 0;
}


void test_unit_ready()
{
    int i;
    __u8 test_unit_cmd[6];

    memset(test_unit_cmd, 0, 6);

/*       */
    for(i = 0; i < 20; i++) {
	if(send_cmd(test_unit_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) == 0) return;
	sleep(1);
    }

    printf("Unit not ready\n");
    exit(-1);
}


int read_dvd_struct(__u8 ffield)
{
    __u8 read_dvd_struct_cmd[12];
    __u16 alloc_size = 2052;
    __u8 buff[2052];
    __u16 dvd_data_len = 0;
    __u32 Start_DA = 0, End_DA = 0;
    __u32 u_part = 0;

    layer_descr_t *ldesc;

    ldesc = (void *)(buff + 4);

    memset(buff, 0, 2052);
    memset(read_dvd_struct_cmd, 0, 12);
    read_dvd_struct_cmd[0] = 0xAD;
    read_dvd_struct_cmd[7] = ffield; // Format field
    read_dvd_struct_cmd[8] = *((__u8 *)&alloc_size + 1);
    read_dvd_struct_cmd[9] = *((__u8 *)&alloc_size);

    test_unit_ready();
    if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, 2052, 20) < 0) return -1;

    *((__u8 *)&dvd_data_len) = buff[1];
    *((__u8 *)&dvd_data_len + 1) = buff[0];
    dvd_data_len += 2; //     (2 bytes)

    printf("\nFormat field: 0x%X\n", ffield);
    printf("DVD STRUCTURE Data Length: %d\n", dvd_data_len);

    printf("Book type: 0x%.2X\n", ldesc->book_type);
    printf("Disk size: %d\n", ldesc->disk_size);

    Start_DA = __swab32(ldesc->start_psn_da);
    End_DA = __swab32(ldesc->end_psn_da);

    printf("Starting PSN of Data Area:  0x%X (%u)\n", Start_DA, Start_DA);
    printf("End PSN of Data Area:  0x%X (%u)\n", End_DA, End_DA);

/* Read unique part (see table #21/22 in INF-8090i) */
    memcpy((void *)&u_part, buff + 4 + 32, 4);
    u_part = __swab32(u_part);
    if(ffield == 0) printf("Start PSN of Current Border-out: ");
    printf("%Xh (%u)\n", u_part, u_part);

    u_part = 0;
    memcpy((void *)&u_part, buff + 4 + 36, 4);
    u_part = __swab32(u_part);
    if(ffield == 0) printf("Start PSN of the next Border-in: ");
    printf("%Xh (%u)\n", u_part, u_part);

    return 0;
}


int main(int argc, char **argv)
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return 1;
    }

/* Read Physical Format Information Zone from Lead-In Area */
    if(read_dvd_struct(0x10) < 0) printf("Cannot read DVD structure\n");

/* Read Physical Format Information Zone from last Border-in */
    if(read_dvd_struct(0x00) < 0) printf("Cannot read DVD structure\n");

    close(sg_fd);

    return 0;
}
