#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/cdrom.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>

#define SG_DEV "/dev/sg0"

int sg_fd;

int send_cmd(__u8 *cmd, __u8 cmdlen, int direction, __u8 *data, \
	    __u32 datalen, unsigned int timeout)
{
    int k = 0;
    sg_io_hdr_t io_hdr;
    __u8 sense_buffer[32];

    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cmdlen;
    io_hdr.mx_sb_len = sizeof(sense_buffer);
    io_hdr.dxfer_direction = direction; //   
    io_hdr.dxfer_len = datalen; //  
    io_hdr.dxferp = data; //    
    io_hdr.cmdp = cmd;
    io_hdr.sbp = sense_buffer;
    io_hdr.timeout = timeout * 1000; /* 20000 millisecs == 20 seconds */

    if(ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
    	perror("SG_IO ioctl");
    	return -1;
    }

    /* now for the error processing */
    if((io_hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK) return -1;

    return 0;
}


void test_unit_ready()
{
    int i;
    __u8 test_unit_cmd[6];

    memset(test_unit_cmd, 0, 6);

/*       */
    for(i = 0; i < 20; i++) {
	if(send_cmd(test_unit_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) == 0) return;
	sleep(1);
    }

    printf("Unit not ready\n");
    exit(-1);
}


int read_dvd_struct()
{

#define SIZE 32776 // 32768 RMD data (1 ECC block) + 8 header`s bytes

    int i, n, k;
    __u8 read_dvd_struct_cmd[12];
    __u8 buff[SIZE];
    __u8 *RMD = buff + 8; //    RMD 

    __u16 alloc_size = SIZE;
    __u16 dvd_data_len = 0;
    __u16 Inv_RZone_num = 0; // Invisible RZone Number (Last Rzone Number)

    __u32 LRS = 0; // Last Record RMA Sector
    __u32 addr; //  Address  
    __u32 Start_BOUT = 0, Start_RZone = 0, LRA_RZone = 0;

/* Get Last Recorded RMA Sector */
    memset(buff, 0, SIZE);
    memset(read_dvd_struct_cmd, 0, 12);
    read_dvd_struct_cmd[0] = 0xAD;
    read_dvd_struct_cmd[7] = 0x0D; // read RMA
    read_dvd_struct_cmd[9] = 8;

    test_unit_ready();
    if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 20) < 0) return -1;

    *((__u8 *)&dvd_data_len) = buff[1];
    *((__u8 *)&dvd_data_len + 1) = buff[0];
    dvd_data_len += 2; //     (2 bytes)
    printf("DVD STRUCTURE Data Length: %d\n", dvd_data_len);

    memcpy((void *)&LRS, buff + 4, 4);
    LRS = __swab32(LRS);
    printf("Last recorded RMA sector number: %u\n", LRS);

    memset(buff, 0, SIZE);
    read_dvd_struct_cmd[8] = *((__u8 *)&alloc_size + 1);
    read_dvd_struct_cmd[9] = *((__u8 *)&alloc_size);

/*   Field #4  RMD . Field #4 -  5-    RMD 
 *  Format #1 ( Linking Loss area)
 */
    for(k = 0;;k++) {

	addr = 16 * k + 5;
	if(addr > LRS) break;

	read_dvd_struct_cmd[2] = *((__u8 *)&addr + 3);
	read_dvd_struct_cmd[3] = *((__u8 *)&addr + 2);
	read_dvd_struct_cmd[4] = *((__u8 *)&addr + 1);
	read_dvd_struct_cmd[5] = *((__u8 *)&addr);

	test_unit_ready();
	if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 20) < 0) return -1;

	printf("\nSECTOR %u:\n", addr);
	*((__u8 *)&Inv_RZone_num + 1) = RMD[0];
	*((__u8 *)&Inv_RZone_num) = RMD[1];
	printf("Invisible/Incomplete RZone Number: %d\n", Inv_RZone_num);

	printf("RZone\tSTART\tLAST\n");
	for(i = 1, n = 16; i <= Inv_RZone_num; i++, n += 8) {

	    memcpy((void *)&Start_RZone, RMD + n, 4);
	    Start_RZone = __swab32(Start_RZone);

	    memcpy((void *)&LRA_RZone, RMD + n + 4, 4);
	    LRA_RZone = __swab32(LRA_RZone);

	    printf("%-d\t%-d\t%-d\n", i, Start_RZone, LRA_RZone);
	}
    }
    printf("\n");
    return 0;
}


int main(int argc, char **argv)
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return -1;
    }

    if(read_dvd_struct() < 0) printf("Cannot read DVD structure\n");

    close(sg_fd);

    return 0;
}
