/*  RMA  DVD-RW ,    INCREMENTAL */

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/cdrom.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>

#define SG_DEV "/dev/sg1"

int sg_fd;

int send_cmd(__u8 *cmd, __u8 cmdlen, int direction, __u8 *data, \
	    __u32 datalen, unsigned int timeout)
{
    int k = 0;
    sg_io_hdr_t io_hdr;
    __u8 sense_buffer[32];

    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cmdlen;
    io_hdr.mx_sb_len = sizeof(sense_buffer);
    io_hdr.dxfer_direction = direction; //   
    io_hdr.dxfer_len = datalen; //  
    io_hdr.dxferp = data; //    
    io_hdr.cmdp = cmd;
    io_hdr.sbp = sense_buffer;
    io_hdr.timeout = timeout * 1000; /* 20000 millisecs == 20 seconds */

    if(ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
    	perror("SG_IO ioctl");
    	return -1;
    }

    /* now for the error processing */
    if((io_hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK) return -1;

    return 0;
}


void test_unit_ready()
{
    int i;
    __u8 test_unit_cmd[6];

    memset(test_unit_cmd, 0, 6);

/*       */
    for(i = 0; i < 20; i++) {
	if(send_cmd(test_unit_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) == 0) return;
	sleep(1);
    }

    printf("Unit not ready\n");
    exit(-1);
}


int read_RMA()
{

#define SIZE 6152 // 2048 * 3 (3 : Field 0, 1, 2)  8  
#define SIZE1 2056 // 2048 (1 )  8   

    int i = 0, n = 0, k = 0;
    __u8 read_dvd_struct_cmd[12];
    __u8 buff[SIZE];
    __u8 *RMD = buff + 8; //    RMD 
    __u8 disk_status = 0;
    __u8 erase_code = 0;

    __u16 alloc_size = SIZE;
    __u16 EOC = 0; // Erase Operation Count
    __u16 RMD_Format = 0; // from  RMD Header (Field 0)
    __u16 Inv_RZone_num = 0; // Invisible RZone Number (Last Rzone Number)

    __u32 addr; //  Address  
    __u32 LRS = 0; // Last Recorded RMA sector number
    __u32 F3_RMD_Set_Pointer = 0; // Format 3 RMD Set Pointer
    __u32 Update_Counter = 0; // From Format 2 RMD Field 1
    __u32 EI1 = 0, EI2 = 0; // Erase Information 1/2 (form Format 2 RMD Field 2, Erase Operation Information)
    __u32 Start_RZone = 0; // Start Sector Number of RZone #n
    __u32 LRA_RZone = 0; // Last Recorded Address of RZone #n

    memset(buff, 0, SIZE);

/*    RMD ,    Format 2:
 * - Field #0 (RMD Header)  RMD Set        RMD ;
 * - Field #1  .   , , Format 3 RMD Set Pointer.   = 2 (Linking Loss Area + Field 0)
 * - Field #2 RMD  .        c .
 *   = 3 (Linking Loss Area + Field #0 + Field #1)
 */
    memset(read_dvd_struct_cmd, 0, 12);
    read_dvd_struct_cmd[0] = 0xAD;
    //read_dvd_struct_cmd[5] = 1; //    
    read_dvd_struct_cmd[7] = 0x0D; // read RMA
    read_dvd_struct_cmd[8] = *((__u8 *)&alloc_size + 1);
    read_dvd_struct_cmd[9] = *((__u8 *)&alloc_size);

    for(i = 0; i < 1 /* 5 */; i++) {

	read_dvd_struct_cmd[5] = i * 16 + 1; //    

	memset(buff, 0, SIZE);
	test_unit_ready();
	if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 20) < 0) return -1;

    /* Get Last Recorded RMA Sector */
	memcpy((void *)&LRS, buff + 4, 4);
	LRS = __swab32(LRS);
	printf("Last recorded RMA sector number: %u\n", LRS);

    /* Field #0,   RMD .     0, 0-  -  Linking Loss area */
	*((__u8 *)&RMD_Format) = RMD[1];
	*((__u8 *)&RMD_Format + 1) = RMD[0];

	disk_status = RMD[2];

	printf("RMD Format: %d\n", RMD_Format);
	printf("Disk Status: %.2Xh\n", disk_status);

    /* Field #1, 2-   RMD */
	memcpy((void *)&Update_Counter, RMD + 0x800, 4);
	Update_Counter = __swab32(Update_Counter);

	memcpy((void *)&F3_RMD_Set_Pointer, (RMD + 4 + 0x800), 4);
	F3_RMD_Set_Pointer = __swab32(F3_RMD_Set_Pointer);

	*((__u8 *)&EOC) = *(RMD + 13 + 0x800);
	*((__u8 *)&EOC + 1) = *(RMD + 12 + 0x800);

	printf("Update Counter: %u\n", Update_Counter);
	printf("Format 3 RMD Set Pointer: %u\n", F3_RMD_Set_Pointer);
	printf("Erase Operation Counter: %d\n", EOC);

    /* Field #2, 3-   RMD */
	erase_code  = RMD[0 + 0x1000];

	memcpy((void *)&EI1, RMD + 2 + 0x1000, 4);
	EI1 = __swab32(EI1);

	memcpy((void *)&EI2, RMD + 6 + 0x1000, 4);
	EI2 = __swab32(EI2);

	printf("Erase Operation Code: %d\n", erase_code);
	printf("Erase Information 1: %u\n", EI1);
	printf("Erase Information 2: %u\n\n", EI2);
    }


/* C   RZone */
    alloc_size = SIZE1; //     
    memset(buff, 0, SIZE);
    read_dvd_struct_cmd[8] = *((__u8 *)&alloc_size + 1);
    read_dvd_struct_cmd[9] = *((__u8 *)&alloc_size);

    for(k = 5;;k++) {

/*    RMD ,     Format 2,
 *    Field #0  Field #4    RMD   Format 1.
 * Field #0 -  1    RMD ,  Field #4 - 5    RMD 
 *  Format 1 ( Linking Loss area)
 */
	addr = 16 * k + 1;
	if(addr > LRS) break;

	read_dvd_struct_cmd[2] = *((__u8 *)&addr + 3);
	read_dvd_struct_cmd[3] = *((__u8 *)&addr + 2);
	read_dvd_struct_cmd[4] = *((__u8 *)&addr + 1);
	read_dvd_struct_cmd[5] = *((__u8 *)&addr);

	test_unit_ready();
	if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 200) < 0) return -1;

	*((__u8 *)&RMD_Format + 1) = RMD[0];
	*((__u8 *)&RMD_Format) = RMD[1];

	disk_status = RMD[2];

	printf("\nField #0 (sector %u)\n", addr);
	printf("RMD Format: %d\t", RMD_Format);
	printf("Disk Status: %.2X\n", disk_status);

/*  Field #4 */

	//addr = 16 * k + 5;
	//if(addr > LRS) break;
	addr += 4;

	read_dvd_struct_cmd[2] = *((__u8 *)&addr + 3);
	read_dvd_struct_cmd[3] = *((__u8 *)&addr + 2);
	read_dvd_struct_cmd[4] = *((__u8 *)&addr + 1);
	read_dvd_struct_cmd[5] = *((__u8 *)&addr);

	test_unit_ready();
	if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 200) < 0) return -1;

	printf("\nField #4 (sector %u)\n", addr);
	*((__u8 *)&Inv_RZone_num + 1) = RMD[0];
	*((__u8 *)&Inv_RZone_num) = RMD[1];
	printf("Invisible/Incomplete RZone Number: %d\n", Inv_RZone_num);

	printf("RZone\tSTART\tLAST\n");
	for(i = 1, n = 16; i <= Inv_RZone_num; i++, n += 8) {

	    memcpy((void *)&Start_RZone, RMD + n, 4);
	    Start_RZone = __swab32(Start_RZone);

	    memcpy((void *)&LRA_RZone, RMD + n + 4, 4);
	    LRA_RZone = __swab32(LRA_RZone);

	    printf("%-d\t%-d\t%-d\n", i, Start_RZone, LRA_RZone);
	}
    }

    printf("\n");
    return 0;
}


int main(int argc, char **argv)
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return -1;
    }

    if(read_RMA() < 0) printf("Cannot read DVD structure\n");

    close(sg_fd);

    return 0;
}
