#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/cdrom.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>

#define SG_DEV "/dev/sg1"

int sg_fd;

int send_cmd(__u8 *cmd, __u8 cmdlen, int direction, __u8 *data, \
	    __u32 datalen, unsigned int timeout)
{
    int k = 0;
    sg_io_hdr_t io_hdr;
    __u8 sense_buffer[32];

    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cmdlen;
    io_hdr.mx_sb_len = sizeof(sense_buffer);
    io_hdr.dxfer_direction = direction; //   
    io_hdr.dxfer_len = datalen; //  
    io_hdr.dxferp = data; //    
    io_hdr.cmdp = cmd;
    io_hdr.sbp = sense_buffer;
    io_hdr.timeout = timeout * 1000; /* 20000 millisecs == 20 seconds */

    if(ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
    	perror("SG_IO ioctl");
    	return -1;
    }

    /* now for the error processing */
    if((io_hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK) return -1;

    return 0;
}


void test_unit_ready()
{
    int i;
    __u8 test_unit_cmd[6];

    memset(test_unit_cmd, 0, 6);

/*       */
    for(i = 0; i < 20; i++) {
	if(send_cmd(test_unit_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) == 0) return;
	sleep(1);
    }

    printf("Unit not ready\n");
    exit(-1);
}


int read_dvd_struct()
{

#define SIZE 2052 // 2048 bytes (1 sector)  + 4 header`s bytes

    int i, n = 16;
    __u8 read_dvd_struct_cmd[12];
    __u8 buff[SIZE];
    __u8 *RMD = buff + 4; //    RMD 
    __u8 disk_status = 0;

    __u16 alloc_size = SIZE;
    __u16 Inv_RZone_num = 0; // Invisible RZone Number (Last Rzone Number)

    __u32 Start_BOUT = 0, Start_RZone = 0, LRA_RZone = 0;

/* Read RMD #0, check Disk status */
    memset(buff, 0, SIZE);
    memset(read_dvd_struct_cmd, 0, 12);
    read_dvd_struct_cmd[0] = 0xAD;
    read_dvd_struct_cmd[7] = 0x0C; // read RMD in the last Border-out
    read_dvd_struct_cmd[5] = 0; // Address = 0, read RMD #0
    read_dvd_struct_cmd[8] = *((__u8 *)&alloc_size + 1);
    read_dvd_struct_cmd[9] = *((__u8 *)&alloc_size);

    test_unit_ready();
    if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 20) < 0) return -1;

    disk_status = *(RMD + 2);
    printf("Disk status: %d\n", disk_status);

/* Read RMD #3, Border Zone Information */
    memset(buff, 0, SIZE);
    read_dvd_struct_cmd[5] = 3; // Address = 3, read RMD #3

    test_unit_ready();
    if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 20) < 0) return -1;

    for(i = 0; i <= 2044; i += 4) {
	memcpy((void *)&Start_BOUT, RMD + i, 4);
	if(!Start_BOUT) break;
	Start_BOUT = __swab32(Start_BOUT);
	printf("Start Sector Number of Border-out #%d: %u\n", i/4 + 1, Start_BOUT);
    }

/* Read RMD #4, RZone Information */
    memset(buff, 0, SIZE);
    read_dvd_struct_cmd[5] = 4; // Address = 4, read RMD #4

    test_unit_ready();
    if(send_cmd(read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, alloc_size, 20) < 0) return -1;

    *((__u8 *)&Inv_RZone_num + 1) = RMD[0];
    *((__u8 *)&Inv_RZone_num) = RMD[1];
    printf("Invisible/Incomplete RZone Number: %d\n", Inv_RZone_num);

    printf("RZone\tSTART\tLAST\n");
    for(i = 1; i <= Inv_RZone_num; i++, n += 8) {

	memcpy((void *)&Start_RZone, RMD + n, 4);
	Start_RZone = __swab32(Start_RZone);

	memcpy((void *)&LRA_RZone, RMD + n + 4, 4);
	LRA_RZone = __swab32(LRA_RZone);

	printf("%-d\t%-d\t%-d\n", i, Start_RZone, LRA_RZone);
    }
    return 0;
}


int main(int argc, char **argv)
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return -1;
    }

    if(read_dvd_struct() < 0) printf("Cannot read DVD structure\n");

    close(sg_fd);

    return 0;
}
