#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>

#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64
#endif

#include <sys/stat.h>
#include <linux/cdrom.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include "iso2dvd.h"

#define SG_DEV "/dev/sg1" //   
int sg_fd; //   

__u16 page5_len;
__u8 *page5_data;
disk_info_block_t *DIB;
wpm_t *wpm; //    

int mode_sense()
{
    __u8 mode_sense_cmd[10];

/*      */
    page5_data = (__u8 *)malloc(2);
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = 2;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, 2, 200) < 0) return -1;

    *((__u8 *)&page5_len) = page5_data[1];
    *((__u8 *)&page5_len + 1) = page5_data[0];
    page5_len += 2;

    free(page5_data);
    page5_data = (__u8 *)malloc(page5_len);

    memset(page5_data, 0, page5_len);
    wpm = (void *)(page5_data + 8); //       

    /* Prepare MODE_SENSE_10 command */
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_sense_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, page5_len, 200) < 0) return -1;

    return 0;
}


int mode_select()
{
    __u8 mode_select_cmd[10];

    if(mode_sense() < 0) {
	printf("-ERR: mode sense\n");
	exit(-1);
    }

    wpm->write_type = 2; //  DAO for DVD
    wpm->test_write = 0; // 0 -    , 1 -    

    /* Prepare MODE_SELECT_10 command */
    memset(mode_select_cmd, 0, 10);
    mode_select_cmd[0] = MODE_SELECT_10;
    mode_select_cmd[1] = 0x10;
    mode_select_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_select_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, mode_select_cmd, 10, SG_DXFER_TO_DEV, page5_data, page5_len, 200) < 0) return -1;

/*      */
    if(mode_sense() < 0) {
	printf("-ERR: mode sense\n");
	exit(-1);
    }

    if(wpm->write_type != 2) {
	printf("-ERR: mode select\n");
	exit(-1);
    }

    free(page5_data);
    return 0;
}


/*     */
int inquiry()
{

#define INQ_SIZE 40

    __u8 inq_cmd[6];
    __u8 data_buff[INQ_SIZE];

    memset(inq_cmd, 0, 6);
    memset(data_buff, 0, INQ_SIZE);
    inq_cmd[0] = INQUIRY;
    inq_cmd[4] = INQ_SIZE; //   

    if(send_cmd(sg_fd, inq_cmd, 6, SG_DXFER_FROM_DEV, data_buff, INQ_SIZE, 20) < 0) return -1;

    printf("\n%.8s%.16s%.4s\n", data_buff + 8, data_buff + 16, data_buff + 32);

    return 0;
}


int read_dvd_struct()
{
    int i;
    __u8 read_dvd_struct_cmd[12];
    __u16 alloc_size = 2052;
    __u8 buff[2052];
    __u16 dvd_data_len = 0;
    layer_descr_t *ldesc = (void *)(buff + 4);

#define BOOK_TYPE (ldesc->book_type) //     DVD-

    test_unit_ready(sg_fd);

    memset(buff, 0, 2052);
    memset(read_dvd_struct_cmd, 0, 12);
    read_dvd_struct_cmd[0] = 0xAD;
    read_dvd_struct_cmd[7] = 0x00;
    read_dvd_struct_cmd[8] = *((__u8 *)&alloc_size + 1);
    read_dvd_struct_cmd[9] = *((__u8 *)&alloc_size);

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, read_dvd_struct_cmd, 12, SG_DXFER_FROM_DEV, buff, 2052, 200) < 0) return -1;

    for(i = 0; i < 6; i++) {
	if(BOOK_TYPE == disk_types[i].dtype) break;
    }
    printf("\nBook Type: %s\n", disk_types[i].dname);

    if(i == 6) {
	printf("Unknown book type\n");
	return -1;
    }

    return 0;
}


/*    DVD-R/-RW */
int check_feature(__u16 f_num)
{

#define FEATURE_LEN 16

    int i;
    __u8 get_conf_cmd[10]; // CDB
    __u8 data_buff[FEATURE_LEN]; //  
    __u32 data_length = 0; //   
    __u16 cur_prof = 0; //  , Current Profile

    dvd_t *dvd;

    memset(data_buff, 0, sizeof(data_buff));
    dvd = (void *)(data_buff + 8);

    memset(get_conf_cmd, 0, 10);
    get_conf_cmd[0] = 0x46; // GET CONFIGURATION
    get_conf_cmd[1] = 2; // RT = 10b
    get_conf_cmd[2] = *((__u8 *)&f_num + 1);
    get_conf_cmd[3] = *((__u8 *)&f_num);
    get_conf_cmd[8] = FEATURE_LEN;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, get_conf_cmd, 10, SG_DXFER_FROM_DEV, data_buff, FEATURE_LEN, 200) < 0) {
	printf("-ERR: check feature\n");
	return -1;
    }

    memcpy((void *)&data_length, data_buff, 4);
    data_length = __swab32(data_length);
    if(data_length == 4) return -1; //   

/*   */
    *((__u8 *)&cur_prof) = data_buff[7];
    *((__u8 *)&cur_prof + 1) = data_buff[6];
    //printf("\nCurrent profile: 0x%04X\n", cur_prof);

    if((dvd->dvd_rw == 1) && (dvd->current == 1)) return 0;

    return -1;
}


/* /  */
int prevent_allow(__u8 prevent_flag)
{
    __u8 prevent_allow_cmd[6];

    memset(prevent_allow_cmd, 0, 6);
    prevent_allow_cmd[0] = 0x1E;
    prevent_allow_cmd[4] = prevent_flag;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, prevent_allow_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) < 0) return -1;

    return 0;
}


/*   */
void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);

    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) < 0) printf("-ERR: eject\n");

    return;
}


int sync_cache()
{
    __u8 flush_cache[10];

    memset(flush_cache, 0, 10);
    flush_cache[0] = 0x35;

    if(send_cmd(sg_fd, flush_cache, 10, SG_DXFER_NONE, NULL, 0, 1000) < 0) return -1;

    printf("OK\n");
    return 0;
}


/*       */
int reserv_track(__u32 track_size)
{
    __u8 reserv_track_cmd[10];
    __u32 size = 0;

    memset(reserv_track_cmd, 0, 10);
    reserv_track_cmd[0] = 0x53;

    size = __swab32(track_size);
    memcpy((void *)(reserv_track_cmd + 5), (void *)&size, 4);

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, reserv_track_cmd, 10, SG_DXFER_NONE, NULL, 0, 20) < 0) return -1;

    return 0;
}


/*     */
int write_iso(__u8 *file_name)
{
    int in_f;
    __u8 write_cmd[10];
    __u8 write_buff[BUFF_SIZE];
    __u32 lba1 = 0, lba = 0;
    ssize_t bsize = 0;

    test_unit_ready(sg_fd);

    in_f = open(file_name, O_RDONLY);
    if(in_f < 0) { perror("open"); exit(errno); }

    memset(write_buff, 0, BUFF_SIZE);

    /* Prepare WRITE_10 command */
    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = SECT_NUM;

    for(;;) {

	bsize = read(in_f, write_buff, BUFF_SIZE);

	if(bsize < 0) { perror("read"); exit(errno); }

	if(bsize == 0) {
	    fprintf(stdout,"\rLSN: %6d\n", lba1); fflush(stdout);
	    return 0;
	}

	if(bsize < BUFF_SIZE) break;

	fprintf(stdout,"\rLSN: %6d", lba1); fflush(stdout);
	lba = __swab32(lba1);
	memcpy((write_cmd + 2), (void *)&lba, 4);

	lba1 += SECT_NUM;

	if(send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, BUFF_SIZE, 200) < 0) return -1;
    }

    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = bsize / 2048;
    fprintf(stdout,"\rLSN: %6d", lba1); fflush(stdout);

    lba = __swab32(lba1);
    memcpy((write_cmd + 2), (void *)&lba, 4);

    if(send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, bsize, 200) < 0) return -1;

    fprintf(stdout,"\rLSN: %6d\n", lba1 + write_cmd[8]);
    fflush(stdout);

    return 0;
}


/*     */
void do_record(__u8 *image_name)
{
    struct stat64 s;
    __u32 track_size; //  

    if(DIB->disk_status != 0) {
	printf("-ERR: not blank disk\n");
	return;
    }

    memset((void *)&s, 0, sizeof(struct stat));
    if(stat64(image_name, &s)) { perror("stat"); return; }

    track_size = (__u32)(s.st_size/2048);

    printf("Image size %s: %u b", image_name, s.st_size/1024);
    printf(" (%u blocks)\n", track_size);

/*     */
    if(mode_select() < 0) return;

/*     */
    if(reserv_track(track_size) < 0) {
        printf("-ERR: reserve track\n");
        return;
    }

/*   */
    prevent_allow(LOCK);

    printf("Record start at %s\n", __TIME__);

    if(write_iso(image_name) < 0) {
	printf("-ERR: write %s\n", image_name);
	return;
    }

    printf("Fixating..\n");
    if(sync_cache() < 0) printf("-ERR: syncing cache\n");
    else printf("Record stop at %s\n", __TIME__);

/*   */
    prevent_allow(UNLOCK);

    eject_cd(); //  

    return;
}


int read_disk_info()
{
    __u8 read_disk_info_cmd[10];

    memset(DIB, 0, DIB_LEN);
    memset(read_disk_info_cmd, 0, 10);
    read_disk_info_cmd[0] = 0x51;
    read_disk_info_cmd[8] = DIB_LEN;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, read_disk_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)DIB, DIB_LEN, 20) < 0) {
	printf("-ERR: read disk info\n");
	return -1;
    }

    return 0;
}


int main(int argc, char **argv)
{
    if(argc != 2) {
	printf("\n\tUsage: iso2dvd [ISO-image]\n\n");
	goto out;
    }

    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
	perror("open");
	exit(-1);
    }

    if(inquiry() < 0) {
	printf("-ERR: inquiry\n");
	goto out;
    }

/*    DVD-R/RW  */
    if(check_feature(DVD_RW) < 0) goto out;

    DIB = (disk_info_block_t *)malloc(DIB_LEN);
    if(DIB == NULL) { perror("malloc"); exit(errno); }

/*     */
    if(read_disk_info() < 0) goto out;

    if(read_dvd_struct() < 0) {
	printf("-ERR: read dvd struct\n");
	goto out;
    }

    do_record(argv[1]);

out:
    //printf("\nFor test only :))\n");
    free(DIB);

    close(sg_fd);

    return 0;
}
