#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>

#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64
#endif

#include <sys/stat.h>
#include <linux/cdrom.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include "iso2dvd_multi.h"

#define SG_DEV "/dev/sg1"

int sg_fd;
__u16 page5_len;
__u8 *page5_data;
wpm_t *wpm; //    


int mode_sense()
{
    __u8 mode_sense_cmd[10];

/*      */
    page5_data = (__u8 *)malloc(2);
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = 2;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, 2, 200) < 0) return -1;

    *((__u8 *)&page5_len) = page5_data[1];
    *((__u8 *)&page5_len + 1) = page5_data[0];
    page5_len += 2;

    free(page5_data);
    page5_data = (__u8 *)malloc(page5_len);

    memset(page5_data, 0, page5_len);
    wpm = (void *)(page5_data + 8); //       

    /* Prepare MODE_SENSE_10 command */
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_sense_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, page5_len, 200) < 0) return -1;
/*
    printf("Page code: %d\n", wpm->page_code);
    printf("Page length: %d (0x%X)\n", wpm->page_length, wpm->page_length);
    printf("Write type: %d\n", wpm->write_type);
    printf("Link size: %d\n", wpm->link_size);
    printf("LS_V: %d\n", wpm->ls_v);
    printf("Multisession: %d\n", wpm->multises);
    printf("Packet size: %u\n\n", __swab32(wpm->packet_size));
*/
    return 0;
}


int mode_select()
{
    __u8 mode_select_cmd[10];

    if(mode_sense() < 0) {
	printf("Mode sense error\n");
	exit(-1);
    }

    wpm->multises = 3;
    wpm->write_type = 0; // Incremental mode
    wpm->packet_size = 0;
    *((__u8 *)&wpm->packet_size + 3) = 16;
    wpm->ls_v = 1;
    wpm->link_size = 16;
    wpm->test_write = 0; // 0 - test write off, 1 - test mode on

    /* Prepare MODE_SELECT_10 command */
    memset(mode_select_cmd, 0, 10);
    mode_select_cmd[0] = MODE_SELECT_10;
    mode_select_cmd[1] = 0x10;
    mode_select_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_select_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, mode_select_cmd, 10, SG_DXFER_TO_DEV, page5_data, page5_len, 200) < 0) return -1;

    if(mode_sense() < 0) {
	printf("Mode sense error\n");
	exit(-1);
    }

    free(page5_data);
    return 0;
}


/*     */
int inquiry()
{

#define INQ_SIZE 40

    __u8 inq_cmd[6];
    __u8 data_buff[INQ_SIZE];

    memset(inq_cmd, 0, 6);
    memset(data_buff, 0, INQ_SIZE);
    inq_cmd[0] = INQUIRY;
    inq_cmd[4] = INQ_SIZE; //   

    if(send_cmd(sg_fd, inq_cmd, 6, SG_DXFER_FROM_DEV, data_buff, INQ_SIZE, 20) < 0) return -1;

    printf("\n%.8s%.16s%.4s\n", data_buff + 8, data_buff + 16, data_buff + 32);

    return 0;
}


int write_iso(__u8 *file_name, __u32 start_lba)
{
    int in_f;
    __u8 write_cmd[10];
    __u8 write_buff[BUFF_SIZE];
    __u32 lba = 0, lba1 = 0;
    ssize_t bsize = 0;

    in_f = open(file_name, O_RDONLY);
    if(in_f < 0) { perror("open"); exit(errno); }

    memset(write_buff, 0, BUFF_SIZE);

    lba1 = start_lba;

    /* Prepare WRITE_10 command */
    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = SECT_NUM;

    test_unit_ready(sg_fd);

    for(;;) {

	bsize = read(in_f, write_buff, BUFF_SIZE);

	if(bsize < 0) { perror("read"); exit(errno); }

	if(bsize == 0) {
	    fprintf(stdout,"\rLBA: %6d (%6d)\n", lba1 - start_lba, lba1); fflush(stdout);
	    return 0;
	}

	if(bsize < BUFF_SIZE) break;

	fprintf(stdout,"\rLBA: %6d (%6d)", lba1 - start_lba, lba1); fflush(stdout);
	lba = __swab32(lba1);
	memcpy((write_cmd + 2), (void *)&lba, 4);
	lba1 += SECT_NUM;

	if(send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, BUFF_SIZE, 200) < 0) return -1;
    }

    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = bsize / CD_FRAMESIZE;

    fprintf(stdout,"\rLBA: %6d (%6d)", lba1 - start_lba, lba1); fflush(stdout);
    lba = __swab32(lba1);
    memcpy((write_cmd + 2), (void *)&lba, 4);

    if(send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, bsize, 200) < 0) return -1;

    fprintf(stdout,"\rLBA: %6d (%6d)\n", lba1 + write_cmd[8] - start_lba, lba1 + write_cmd[8]);
    fflush(stdout);

    return 0;
}


int reserv_track(__u32 track_size)
{
    __u8 reserv_track_cmd[10];
    __u32 size = 0;

    memset(reserv_track_cmd, 0, 10);
    reserv_track_cmd[0] = 0x53;

    size = __swab32(track_size);
    memcpy((void *)(reserv_track_cmd + 5), (void *)&size, 4);

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, reserv_track_cmd, 10, SG_DXFER_NONE, NULL, 0, 20) < 0) return -1;

    return 0;
}


__u32 read_track_info(int trk_num)
{
    __u8 read_track_info_cmd[10];
    __u8 data_buff[40];
    __u32 nwa = 0;

    memset(data_buff, 0, 40);
    memset(read_track_info_cmd, 0, 10);
    read_track_info_cmd[0] = 0x52;
    read_track_info_cmd[1] = 1;
    read_track_info_cmd[5] = trk_num;
    read_track_info_cmd[7] = 40;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, read_track_info_cmd, 10, SG_DXFER_FROM_DEV, data_buff, 40, 20) < 0) exit(-1);

    memcpy((void *)&nwa, (void *)(data_buff + 8), 4);

    return __swab32(nwa);
}


int sync_cache()
{
    __u8 flush_cache[10];

    memset(flush_cache, 0, 10);
    flush_cache[0] = SYNCHRONIZE_CACHE;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, flush_cache, 10, SG_DXFER_NONE, NULL, 0, 1000) < 0) return -1;

    return 0;
}


int close_track(int trk_num)
{
    __u8 close_trk_cmd[10];

    if(sync_cache() < 0) printf("\n  \n");

    memset(close_trk_cmd, 0, 10);
    close_trk_cmd[0] = 0x5B;
    close_trk_cmd[2] = 1;
    close_trk_cmd[5] = trk_num; //  

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, close_trk_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    printf("OK\n");
    return 0;
}


int close_sess()
{
    __u8 close_sess_cmd[10];

    memset(close_sess_cmd, 0, 10);
    close_sess_cmd[0] = 0x5B;
    close_sess_cmd[2] = 2;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, close_sess_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


/* /  */
int prevent_allow(__u8 prevent_flag)
{
    __u8 prevent_allow_cmd[6];

    memset(prevent_allow_cmd, 0, 6);
    prevent_allow_cmd[0] = 0x1E;
    prevent_allow_cmd[4] = prevent_flag;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, prevent_allow_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) < 0) return -1;

    return 0;
}


void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);
    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    test_unit_ready(sg_fd);
    send_cmd(sg_fd, start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 20);

    printf("OK\n");
    return;
}



int main(int argc, char **argv)
{
    int i = 1;
    int trk_num;
    struct stat64 s;
    __u32 track_size;
    __u32 start_lba = 0;

    if(argc != 3) {
	printf("\n\tUsage: iso2dvd_multi [ISO-image] [track num]\n\n");
	return 0;
    }

    trk_num = atoi(argv[2]);
    printf("Track number: %d\n", trk_num);

    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return 1;
    }

    if(inquiry() < 0) {
	printf("-ERR: inquiry\n");
	return -1;
    }

    if(mode_select() < 0) {
	printf("Mode select error\n");
	return 0;
    }

    memset((void *)&s, 0, sizeof(struct stat));
    if(stat64(argv[1], &s)) {
        perror("stat");
        return -1;
    }

    track_size = (__u32)(s.st_size/2048);

    printf("File size: %u Kbytes\n", s.st_size/1024);
    printf("Sectors in file: %u\n", track_size);

    printf("Reserve track #%d...", trk_num);
    if(reserv_track(track_size) < 0) {
        printf("Cannot reserve track #%d\n", trk_num);
        return -1;
    }
    printf("OK\n");

    start_lba = read_track_info(trk_num);
    printf("Start address: %u\n", start_lba);

    printf("Record start at %s\n", __TIME__);

    prevent_allow(1);

    if(write_iso(argv[1], start_lba) < 0) {
	printf("Cannot burn image %s\n", argv[1]);
	return -1;
    }

    printf("Close track..\n");
    if(close_track(trk_num) < 0) printf("-ERR: close track\n");

    printf("Close session..\n");
    if(close_sess() < 0) printf("-ERR: close session\n\n");
    else printf("OK\n");

    //printf("OK\nRecord stop at %s\n", __TIME__);

    prevent_allow(0);

    eject_cd();

    close(sg_fd);
    return 0;
}
