/*     (TOC) */

#include <stdio.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include <send_cmd.h>

struct toc {
    __u8 res;
    __u8 adr_cntl;
    __u8 trk_num;
    __u8 res1;
    __u32 lba;
    struct toc *next;
} __attribute__ ((packed));

struct toc_entry {
    __u32 start_lba;
    __u32 end_lba;
} __attribute__ ((packed));

int max_track_num = 0;


/*    */
void test_unit_ready()
{
    __u8 cmd_buff[12];
    memset((void *)cmd_buff, 0, 12);
    cmd_buff[0] = 0x00;
    send_packet_command(cmd_buff);

    return;
}


int request_sense()
{
    int i = 0;
    __u8 cmd_buff[12];
    __u8 sense_buff[14];
    __u16 a;

#define SK (sense_buff[2] & 0x0F)
#define ASC sense_buff[12]
#define ASCQ sense_buff[13]

    memset((void *)cmd_buff, 0, 12);
    memset((void *)sense_buff, 0, 14);

    cmd_buff[0] = 0x03;
    cmd_buff[4] = 14;

    if(send_packet_data_command(14, cmd_buff) < 0) {
	fprintf(stderr, "Error request sense\n");
	return -1;
    }

    for(i = 0; i < 14; i += 2) {
        IN_P_W(a, 0x170);
	sense_buff[i] = *((__u8 *)&a);
	sense_buff[i + 1] = *((__u8 *)&a + 1);
    }

    fprintf(stderr, "SK/ASC/ASCQ: 0x%X/0x%X/0x%X\n", SK, ASC, ASCQ);
    fflush(stderr);

    return 0;
}


struct toc * read_toc()
{

#define READ_TOC 0x43

    static int i = 1;
    int n;
    __u8 cmd_buff[12];
    __u8 data_buff[12];
    __u16 a, toc_len = 0;
    struct toc *t;

    if(i > 0xAA) return NULL;

    t = (struct toc *)malloc(sizeof(struct toc));

    memset((void *)cmd_buff, 0, 12);
    cmd_buff[0] = READ_TOC;
    cmd_buff[6] = i;
    cmd_buff[8] = 12;

    if(send_packet_data_command(12, cmd_buff) < 0) {
	printf("Error read TOC\n");
	request_sense();
	exit(-1);
    }

    memset(data_buff, 0, 12);
    for(n = 0; n < 12; n += 2) {
        IN_P_W(a, 0x170);
	data_buff[n] = *((__u8 *)&a);
	data_buff[n + 1] = *((__u8 *)&a + 1);
    }

    if (i == 1) {

	*((__u8 *)&toc_len) = data_buff[1];
	*((__u8 *)&toc_len + 1) = data_buff[0];

	printf("TOC lenght - %d\n", toc_len);
        printf("First: %d\t", data_buff[2]);
        printf("Last: %d\n", data_buff[3]);
	max_track_num = data_buff[3];
    }

    memcpy((void *)t, (data_buff + 4), 8);
    t->lba = __swab32(t->lba);

    i += 1;
    if(i == (max_track_num + 1)) i = 0xAA;

    t->next = read_toc();
    return t;
}


void view_toc(struct toc *t)
{
    if(t == NULL) return;

    if(t->trk_num == 0xAA) printf("lead out:\t");
    else printf("track: %d\t", t->trk_num);
    printf("lba: %u\n", t->lba);

    view_toc(t->next);
}


struct toc_entry * get_toc_entry(int trk_num, struct toc *t)
{
    struct toc_entry *t_entry;
    int i = 1;

    for(;i < trk_num; i++) t = t->next;

    t_entry = (struct toc_entry *)malloc(sizeof(struct toc_entry));
    t_entry->start_lba = t->lba;
    t_entry->end_lba = t->next->lba;

    return t_entry;
}



int main()
{
    int trk_num = 0;
    struct toc *t = NULL;
    struct toc_entry *t_entry;

    ioperm(0x170, 8, 1);

    drive_reset();

    test_unit_ready();

    t = read_toc();
    view_toc(t);

    printf("Enter track number: ");
    scanf("%d", &trk_num);

    if((trk_num < 1) || (trk_num > max_track_num)) {
	printf("Wrong track number\n");
	return -1;
    }

    t_entry = get_toc_entry(trk_num, t);
    printf("Track %d:\n", trk_num);
    printf("Start LBA - %u\n", t_entry->start_lba);
    printf("End LBA - %u\n", t_entry->end_lba);

    ioperm(0x170, 8, 0);

    return 0;
}
