/*     (TOC) */

#include <stdio.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include <send_cmd.h>

int total_tracks = 0;  //    CDROM

struct toc {
    __u8 res;
    __u8 adr_cntl;
    __u8 trk_num;
    __u8 res1;
    __u32 lba;
} __attribute__ ((packed)) *t;


void test_unit_ready()
{
    __u8 cmd_buff[12];
    memset((void *)cmd_buff, 0, 12);
    cmd_buff[0] = 0x00;
    send_packet_command(cmd_buff);

    return;
}


int request_sense()
{
    int i = 0;
    __u8 cmd_buff[12];
    __u8 sense_buff[14];
    __u16 a;

#define SK sense_buff[2] & 0x0F
#define ASC sense_buff[12]
#define ASCQ sense_buff[13]

    memset((void *)cmd_buff, 0, 12);
    memset((void *)sense_buff, 0, 14);

    cmd_buff[0] = 0x3;
    cmd_buff[4] = 14;

    if(send_packet_data_command(14, cmd_buff) < 0) {
	printf("Error request sense\n");
	return -1;
    }

    for(i = 0; i < 14; i += 2) {
        IN_P_W(a, 0x170);
	sense_buff[i] = *((__u8 *)&a);
	sense_buff[i + 1] = *((__u8 *)&a + 1);
    }

    fprintf(stderr, "SK/ASC/ASCQ: 0x%X/0x%X/0x%X\n", SK, ASC, ASCQ);
    fflush(stderr);

    return 0;
}



int read_toc()
{

#define READ_TOC 0x43
#define MAX_TOC_LEN 804

    int i = 0;
    __u8 cmd_buff[12];
    __u8 *data_buff;
    __u16 a, toc_len = 0; // toc_len -  TOC

    memset((void *)cmd_buff, 0, 12);
    cmd_buff[0] = READ_TOC;
    cmd_buff[6] = 0;

    toc_len = MAX_TOC_LEN;
    cmd_buff[7] = *((__u8 *)&toc_len + 1);
    cmd_buff[8] = *((__u8 *)&toc_len);

    data_buff = (__u8 *)malloc(MAX_TOC_LEN);
    memset(data_buff, 0, MAX_TOC_LEN);

    if(send_packet_data_command(MAX_TOC_LEN, cmd_buff) < 0) {
	request_sense();
	return -1;
    }

    for(i = 0;;i += 2) {
        IN_P_W(a, 0x170);
	if(i == 0) toc_len = __swab16(a);
	if(i > toc_len) break;
	data_buff[i] = *((__u8 *)&a);
	data_buff[i + 1] = *((__u8 *)&a + 1);
    }

    total_tracks = data_buff[3];

    fprintf(stdout,"TOC length: %d\n", toc_len);
    fprintf(stdout,"First: %d\t", data_buff[2]);
    fprintf(stdout,"Last: %d\n", total_tracks);

    t = (struct toc *)malloc(toc_len);
    memcpy((void *)t, (data_buff + 4), toc_len);

    free(data_buff);

    for(i = 0; i < total_tracks; i++)
	fprintf(stdout,"track: %d\tlba: %u\n", (i + 1), __swab32((t + i)->lba));

    fflush(stdout);
    return 0;
}



int main()
{
    int trk_num = 0;

    ioperm(0x170, 8, 1);

    drive_reset();

    test_unit_ready();

    if(read_toc() < 0) {
	printf("Error read TOC\n");
	return -1;
    }

    printf("Enter track number: ");
    scanf("%d", &trk_num);

    if((trk_num < 1) || (trk_num > total_tracks)) {
	printf("Wrong track number\n");
	return -1;
    }

    fprintf(stdout,"track: %d\tlba: %u\n", trk_num, \
	__swab32((t + trk_num - 1)->lba));

    free(t);

    ioperm(0x170, 8, 0);

    return 0;
}
