#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/types.h>

#define SG_DEV "/dev/sg0"
int sg_fd;

/* Disk Information Block (   ) */
typedef struct {
    __u16 disk_info_len;
    __u8 disk_status		:2;
    __u8 state_last_sess	:2;
    __u8 erasable		:1;
    __u8 res			:3;
    __u8 first_track_num;
    __u8 sess_num_lsb;
    __u8 first_trk_num_ls;
    __u8 last_trk_num_ls;
    __u8 bits;
    __u8 disk_type;
    __u8 sess_num_msb;
    __u8 first_trk_num_ms;	// First track number in Last session (Most Sign. Bit)
    __u8 last_trk_num_ms;	// Last track number in Last session (Most Sign. Bit)
    __u32 disk_ident;		// Disk Identification
    __u32 lead_in_start;	// Last session Lead-In Start Address
    __u8 l_out[4];		// Lead-Out possible
    __u8 disk_bar_code[8];
    __u8 disk_appl_code;
    __u8 OPC_table_num;
} disk_info_block_t;

disk_info_block_t *DIB;

int read_disk_info()
{
    __u8 read_disk_info_cmd[10];

    memset(DIB, 0, 34);
    memset(read_disk_info_cmd, 0, 10);
    read_disk_info_cmd[0] = 0x51;
    read_disk_info_cmd[8] = 34;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, read_disk_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)DIB, 34, 200) < 0) return -1;

    return 0;
}


/*     */
int inquiry()
{
    __u8 inq_cmd[6];
    __u8 data_buff[40];

    memset(inq_cmd, 0, 6);
    memset(data_buff, 0, 40);
    inq_cmd[0] = INQUIRY;
    inq_cmd[4] = 40; //   

    if(send_cmd(sg_fd, inq_cmd, 6, SG_DXFER_FROM_DEV, data_buff, 40, 200) < 0) return -1;

    printf("\n%.8s%.16s%.4s\n", data_buff + 8, data_buff + 16, data_buff + 32);

    return 0;
}


/* /  */
int prevent_allow(__u8 prevent_flag)
{
    __u8 prevent_allow_cmd[6];

    memset(prevent_allow_cmd, 0, 6);
    prevent_allow_cmd[0] = 0x1E;
    prevent_allow_cmd[4] = prevent_flag;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, prevent_allow_cmd, 6, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


/*   */
void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);

    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    send_cmd(sg_fd, start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 200);

    return;
}


/*   */
void blank()
{
    int blank_mode;
    __u8 blank_cmd[12];

    if(DIB->erasable != 1) {
	printf("\n D-   .\n\n");
	return;
    }

    printf("\n :\n");
    printf("1.   .\n");
    printf("2.   .\n");
    printf(" : ");
    scanf("%d", &blank_mode);

    if((blank_mode < 1) || (blank_mode > 2)) {
	printf("\n  .\n\n");
	exit(0);
    }

    prevent_allow(1); //  

    memset(blank_cmd, 0, 12);
    blank_cmd[0] = 0xA1; // BLANK
    blank_cmd[1] = blank_mode - 1;

    printf("\n ");
    if(blank_mode == 1) printf(" ");
    else printf(" ");
    printf(" ..\n");

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, blank_cmd, 12, SG_DXFER_NONE, NULL, 0, 9600) < 0) \
    printf("\n     .\n\n");
    else printf("OK\n\n");

    prevent_allow(0);
    eject_cd();

    return;
}


int main()
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
	perror("open");
	return -1;
    }

    if(inquiry() < 0) {
	printf("\n   \n");
	return -1;
    }

    DIB = (disk_info_block_t *)malloc(34);
    if(DIB == NULL) { perror("malloc"); exit(errno); }

    if(read_disk_info() < 0) return -1;

    blank();

    close(sg_fd);

    return 0;
}
