#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include <linux/cdrom.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>

#define SG_DEV "/dev/sg0"

typedef struct {
    __u16 disk_info_len;
    __u8 disk_status		:2;
    __u8 ls_status		:2;
    __u8 erasable		:1;
    __u8 res			:3;
    __u8 first_track_num;
    __u8 sess_num_lsb;
    __u8 first_trk_ls;
    __u8 last_trk_ls;
    __u8 BG_format_status	:2;
    __u8 Dbit			:1;
    __u8 res1			:1;
    __u8 DAC_V			:1;
    __u8 URU			:1;
    __u8 DBC_V			:1;
    __u8 DID_V			:1;
    __u8 disk_type;
    __u8 sess_num_msb;
    __u8 first_trk_ms;
    __u8 last_trk_ms;
    __u32 disk_ident;
    __u32 lead_in_start;
    __u8 l_out[4];
    __u8 disk_bar_code[8];
    __u8 disk_appl_code;
    __u8 OPC_table_num;
} __attribute__ ((packed)) disk_info_block_t;

disk_info_block_t *DIB;
#define DIB_LEN 34

typedef struct {
    __u16 trk_info_len;
    __u8 trk_num_ls;
    __u8 sess_num_ls;
    __u8 res;
    __u8 trk_mode	:4;
    __u8 copy		:1;
    __u8 damage		:1;
    __u8 res1		:2;
    __u8 data_mode	:4;
    __u8 FP		:1;
    __u8 PAC_INC	:1;
    __u8 blank		:1;
    __u8 RT		:1;
    __u8 NWA_V		:1;
    __u8 LRA_V		:1;
    __u8 res2		:6;
    __u32 start_addr;
    __u32 nwa;
    __u32 free_blk;
    __u32 FPS;
    __u32 trk_size;
    __u32 lra;
    __u8 trk_num_ms;
    __u8 sess_num_ms;
    __u16 res3;
} trk_info_t;

trk_info_t *trk_info;
#define TRK_INFO_LEN 36


int sg_fd;


int read_track_info(int trk_num)
{
    __u8 read_track_info_cmd[10];
    __u16 blk_size = 0;

    test_unit_ready(sg_fd);

    memset(read_track_info_cmd, 0, 10);
    read_track_info_cmd[0] = 0x52;
    read_track_info_cmd[1] = 1;
    read_track_info_cmd[5] = trk_num;
    read_track_info_cmd[7] = TRK_INFO_LEN;

    if(send_cmd(sg_fd, read_track_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)trk_info, TRK_INFO_LEN, 200) < 0) {
	printf("Error read info about track #%d\n", trk_num);
	return -1;
    }

    *((__u8 *)&blk_size) = *((__u8 *)&trk_info->trk_info_len + 1);
    *((__u8 *)&blk_size + 1) = *((__u8 *)&trk_info->trk_info_len);

    printf("Track info size: %u\n", blk_size + 2);
    printf("Track number: %d\n", (trk_info->trk_num_ms << 8) | trk_info->trk_num_ls);
    printf("Start address of track: %u\n", __swab32(trk_info->start_addr));
    trk_info->nwa = __swab32(trk_info->nwa);
    printf("Next writable address: %u\n", trk_info->nwa);
    printf("Track size: %u\n", __swab32(trk_info->trk_size));
    printf("Free blocks: %u\n", __swab32(trk_info->free_blk));
    printf("\nRT Blank PAC/INC FP\n%2d%5d%6d%6d\n\n", trk_info->RT, trk_info->blank, trk_info->PAC_INC, trk_info->FP);

    return 0;
}


int read_disk_info()
{
    int i = 0;
    __u8 read_disk_info_cmd[10];
    __u16 disk_info_len = 0;
    __u16 trk_num = 0;

    DIB = (disk_info_block_t *)malloc(DIB_LEN);
    if(DIB == NULL) { perror("malloc"); exit(errno); }

    test_unit_ready(sg_fd);

    memset(DIB, 0, DIB_LEN);
    memset(read_disk_info_cmd, 0, 10);
    read_disk_info_cmd[0] = 0x51;
    read_disk_info_cmd[8] = DIB_LEN;

    if(send_cmd(sg_fd, read_disk_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)DIB, DIB_LEN, 200) < 0) {
	printf("Cannot read disk info!\n");
	return -1;
    }

    *((__u8 *)&disk_info_len) = *((__u8 *)&DIB->disk_info_len + 1);
    *((__u8 *)&disk_info_len + 1) = *((__u8 *)&DIB->disk_info_len);

    trk_num = (DIB->last_trk_ms << 8) | DIB->last_trk_ls;

    printf("Data info length: %d\n", disk_info_len);
    printf("Erasable: %d\n", DIB->erasable);
    printf("Disk status: 0x%.2x\n", DIB->disk_status);
    printf("Number of first track: %d\n", DIB->first_track_num);
    printf("Number of Session: %d\n", (DIB->sess_num_msb << 8) | DIB->sess_num_lsb);
    printf("Status of last Session: 0x%.2x\n", DIB->ls_status);
    printf("First track in last Session: %d\n", (DIB->first_trk_ms << 8) | DIB->first_trk_ls);
    printf("Last track in last Session: %d\n\n", trk_num);

    trk_info = (trk_info_t *)malloc(TRK_INFO_LEN);
    if(trk_info == NULL) { perror("malloc"); exit(errno); }

    if(DIB->disk_status != 0) for(i = 1; i <= trk_num; i++) read_track_info(i);

    free(DIB);
    free(trk_info);

    return 0;
}


int main(int argc, char **argv)
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return -1;
    }

    read_disk_info();

    close(sg_fd);
    return 0;
}
