#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>

#define SG_DEV "/dev/sg0"
#define MSF2LBA(Min, Sec, Frame) (((Min * 60 + Sec) * 75 + Frame) - 150)

int sg_fd; //  

/*     PMA */
typedef struct {
    __u8 rez;		// reserved
    __u8 ctrl	:4;	// Control
    __u8 adr	:4;	// ADR
    __u8 tno;		// TNO (always 0)
    __u8 point;		// POINT
    __u8 min;		// AMIN
    __u8 sec;		// ASEC
    __u8 frame;		// AFRAME
    __u8 zero;		// 0
    __u8 pmin;		// PMIN
    __u8 psec;		// PSEC
    __u8 pframe;	// PFRAME
} __attribute__ ((packed)) pma_t;

int read_pma()
{
    int i, k;
    __u8 read_pma_cmd[10];

    __u8 *pma_data_buff; //      PMA
    __u16 pma_data_len = 0; //    PMA
    __u32 lba;
    int pma_entries = 0; //   PMA
    pma_t *pma;

    test_unit_ready(sg_fd);

    pma_data_buff = (__u8 *)malloc(0xFFFF);
    memset(pma_data_buff, 0, 0xFFFF);

    memset(read_pma_cmd, 0, 10);
    read_pma_cmd[0] = 0x43; // READ_TOC/PMA/ATIP
    read_pma_cmd[2] = 3; // Format Field = 011b, PMA
    read_pma_cmd[7] = 0xFF;
    read_pma_cmd[8] = 0xFF;

    if(send_cmd(sg_fd, read_pma_cmd, 10, SG_DXFER_FROM_DEV, pma_data_buff, 0xFFFF, 200) < 0) return -1;

    *((__u8 *)&pma_data_len) = pma_data_buff[1];
    *((__u8 *)&pma_data_len + 1) = pma_data_buff[0];
    printf("PMA data length: %d\n", pma_data_len + 2);

    pma_entries = (pma_data_len - 2)/11;
    printf("PMA entries: %d\n", pma_entries);

    pma = (pma_t *)malloc(pma_data_len);
    memset((void *)pma, 0, pma_data_len);
    memcpy((void *)pma, pma_data_buff + 4, pma_data_len);

    free(pma_data_buff);

    printf("Entry\tADR\tCTRL\tPoint\tZero\tMin\tSec\tFrame\tPMin\tPsec\tPFrame\tLBA\n");

#define PMIN(i) (pma + i)->pmin
#define PSEC(i) (pma + i)->psec
#define PFRAME(i) (pma + i)->pframe
#define ADR(i) (pma + i)->adr


    for(i = 0; i < pma_entries; i++) {
	printf("%d\t", i);
	printf("%X\t", ADR(i));
	printf("%X\t", (pma + i)->ctrl);
	printf("%X\t", (pma + i)->point);
	printf("%d\t", (pma + i)->zero);
	printf("%d\t", (pma + i)->min);
	printf("%d\t", (pma + i)->sec);
	printf("%d\t", (pma + i)->frame);
	printf("%d\t", PMIN(i));
	printf("%d\t", PSEC(i));
	printf("%d\t", PFRAME(i));

	lba = MSF2LBA(PMIN(i), PSEC(i), PFRAME(i));

	if(ADR(i) != 1) printf("---\n");
	else printf("%u\n", lba);

    }

    free(pma);

    return 0;
}


int main()
{
    sg_fd = open(SG_DEV, O_RDWR);
    if(sg_fd < 0) { perror("open"); return -1; }

    if(read_pma() < 0) printf("Cannot read PMA\n");

    close(sg_fd);
    return 0;
}
