#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <linux/cdrom.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <sys/stat.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include "cue_audio.h"

int sg_fd;
__u16 page5_len;
__u8 *page5_data;
wpm_t *wpm; //    
disk_info_block_t *DIB;

int send_cmd(__u8 *cmd, __u8 cmdlen, int direction, __u8 *data, \
	    __u32 datalen, unsigned int timeout)
{
    sg_io_hdr_t io_hdr;
    __u8 sense_buffer[32];

#define SK (sense_buffer[2] & 0x0F)
#define ASC sense_buffer[12]
#define ASCQ sense_buffer[13]

    memset(&io_hdr, 0, sizeof(sg_io_hdr_t));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = cmdlen;
    io_hdr.mx_sb_len = sizeof(sense_buffer);
    io_hdr.dxfer_direction = direction;
    io_hdr.dxfer_len = datalen; //  
    io_hdr.dxferp = data; //    
    io_hdr.cmdp = cmd;
    io_hdr.sbp = sense_buffer;
    io_hdr.timeout = timeout * 1000; /* 20000 millisecs == 20 seconds */

rep:
    if(ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
    	perror("SG_IO ioctl");
    	return -1;
    }

    if((io_hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK) {
        if(io_hdr.sb_len_wr > 0) {

/*  SK/ASC/ASCQ == 02/04/08 (Not Ready. Long Write in Progress),
 *     WRITE_10
 */
	    if((SK == NOT_READY) && (ASC == 0x04) && (ASCQ == 0x08)) goto rep;
            syslog(LOG_INFO, "Command: 0x%02X", io_hdr.cmdp[0]);
            syslog(LOG_INFO, "Sense data (SK/ASC/ASCQ): 0x%02X/0x%02X/0x%02X", SK, ASC, ASCQ);
        }

        if(io_hdr.masked_status)
            syslog(LOG_INFO, "SCSI status = 0x%x\n", io_hdr.status);
        if(io_hdr.host_status)
	    syslog(LOG_INFO, "Host status = 0x%x\n", io_hdr.host_status);
        if(io_hdr.driver_status)
	    syslog(LOG_INFO, "Driver status = 0x%x\n", io_hdr.driver_status);

	return -1;
    }

    return 0;
}


void test_unit_ready()
{
    int i;
    __u8 test_unit_cmd[6];

    /* Prepare TEST UNIT command */
    memset(test_unit_cmd, 0, 6);

    for(i = 0; i < 50; i++) {
	if(send_cmd(test_unit_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) == 0) return;
	sleep(1);
    }

    printf("\n  .\n");
    exit(-1);
}


int mode_sense()
{
    __u8 mode_sense_cmd[10];

    test_unit_ready();

/*      */
    page5_data = (__u8 *)malloc(2);
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = 2;

    if(send_cmd(mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, 2, 200) < 0) return -1;

    *((__u8 *)&page5_len) = page5_data[1];
    *((__u8 *)&page5_len + 1) = page5_data[0];
    page5_len += 2;

#ifdef DEBUG
    printf("    - %d \n", page5_len);
#endif

    free(page5_data);
    page5_data = (__u8 *)malloc(page5_len);

    memset(page5_data, 0, page5_len);
    wpm = (void *)(page5_data + 8); //       

    /* Prepare MODE_SENSE_10 command */
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = page5_len;

    if(send_cmd(mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, page5_len, 200) < 0) return -1;

    return 0;
}


int mode_select(int test_record_flag)
{
    __u8 mode_select_cmd[10];

    if(mode_sense() < 0) {
	printf("Mode sense error\n");
	exit(-1);
    }

    wpm->write_type = 2; // SAO
    wpm->multises = 0;
    wpm->test_write = test_record_flag;

    test_unit_ready();

    /* Prepare MODE_SELECT_10 command */
    memset(mode_select_cmd, 0, 10);
    mode_select_cmd[0] = MODE_SELECT_10;
    mode_select_cmd[1] = 0x10;
    mode_select_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_select_cmd[8] = *(__u8 *)&page5_len;

    if(send_cmd(mode_select_cmd, 10, SG_DXFER_TO_DEV, page5_data, page5_len, 200) < 0) return -1;

    free(page5_data);
    return 0;
}


/*     */
int inquiry()
{

#define INQ_SIZE 40

    __u8 inq_cmd[6];
    __u8 data_buff[INQ_SIZE];

    memset(inq_cmd, 0, 6);
    memset(data_buff, 0, INQ_SIZE);
    inq_cmd[0] = INQUIRY;
    inq_cmd[4] = INQ_SIZE; //   

    if(send_cmd(inq_cmd, 6, SG_DXFER_FROM_DEV, data_buff, INQ_SIZE, 20) < 0) return -1;

    printf("%.8s%.16s%.4s\n", data_buff + 8, data_buff + 16, data_buff + 32);

    return 0;
}


/*    CD Mastering */
int check_feature(__u16 f_num)
{

#define FEATURE_LEN 16

    int i;
    __u8 get_conf_cmd[10]; // CDB
    __u8 data_buff[FEATURE_LEN]; //  
    __u32 data_length = 0; //   

#ifdef DEBUG
    __u16 cur_prof = 0; //  , Current Profile
#endif

    cd_sao_t *cd_sao;

    test_unit_ready();

    memset(data_buff, 0, sizeof(data_buff));
    cd_sao = (void *)(data_buff + 8);

    memset(get_conf_cmd, 0, 10);
    get_conf_cmd[0] = 0x46; //GET CONFIGURATION
    get_conf_cmd[1] = 2; //RT = 10b
    get_conf_cmd[2] = *((__u8 *)&f_num + 1);
    get_conf_cmd[3] = *((__u8 *)&f_num);
    get_conf_cmd[8] = FEATURE_LEN;

    if(send_cmd(get_conf_cmd, 10, SG_DXFER_FROM_DEV, data_buff, FEATURE_LEN, 20) < 0) {
	printf("   \n\n");
	return -1;
    }

    memcpy((void *)&data_length, data_buff, 4);
    data_length = __swab32(data_length);
    if(data_length == 4) return -1; //   

    if(cd_sao->current != 1) return -1;

#ifdef DEBUG
/*   */
    *((__u8 *)&cur_prof) = data_buff[7];
    *((__u8 *)&cur_prof + 1) = data_buff[6];
    printf("\n  - 0x%04X\n", cur_prof);
#endif

    return (cd_sao->sao);
}


/* /  */
int prevent_allow(__u8 prevent_flag)
{
    __u8 prevent_allow_cmd[6];

    test_unit_ready();

    memset(prevent_allow_cmd, 0, 6);
    prevent_allow_cmd[0] = 0x1E;
    prevent_allow_cmd[4] = prevent_flag;

    if(send_cmd(prevent_allow_cmd, 6, SG_DXFER_NONE, NULL, 0, 20) < 0) return -1;

    return 0;
}


void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);

    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    send_cmd(start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 20);

    return;
}


/*   */
void blank()
{
    int blank_mode;
    __u8 blank_cmd[12];

#define LOCK 1
#define UNLOCK 0

    if(DIB->erasable != 1) {
	printf("\n D-   .\n\n");
	return;
    }

    printf("\n :\n");
    printf("1.   .\n");
    printf("2.   .\n");
    printf(" : ");
    scanf("%d", &blank_mode);

    if((blank_mode < 1) || (blank_mode > 2)) {
	printf("\n  .\n\n");
	exit(0);
    }

    prevent_allow(LOCK); //  

    test_unit_ready();

    memset(blank_cmd, 0, 12);
    blank_cmd[0] = 0xA1; // BLANK
    blank_cmd[1] = blank_mode - 1;

    printf("\n ");
    if(blank_mode == 1) printf(" ");
    else printf(" ");
    printf(" ..\n");

    if(send_cmd(blank_cmd, 12, SG_DXFER_NONE, NULL, 0, 9600) < 0)
    printf("\n     .\n\n");
    else printf("OK\n\n");

    prevent_allow(UNLOCK);
    eject_cd();

    return;
}


int write_audio(__u8 *file_name, int flag)
{

#define PREGAP_SIZE 352800 // 2352 * 150

    int ret, i;
    int in_f;
    __u8 write_cmd[10];
    __u8 *write_buff;
    int lba = 0;
    static int lba1 = -150;
    ssize_t bsize = 0;

    in_f = open(file_name, O_RDONLY);
    if(in_f < 0) { perror("open"); exit(errno); }

/*   WAV- */
    lseek(in_f, 44, 0);

    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;

    test_unit_ready();

    if(flag == 1) { //   

	write_buff = (__u8 *)malloc(PREGAP_SIZE);
	memset(write_buff, 0, PREGAP_SIZE);

	write_cmd[8] = 150; // 0x96

/*  Pre-gap   */
	printf("Writing Pre-Gap..");
        lba = __swab32(lba1);
        memcpy((write_cmd + 2), (void *)&lba, 4);
	lba1 += 150;

        while(1) {
	    ret = send_cmd(write_cmd, 10, SG_DXFER_TO_DEV, write_buff, PREGAP_SIZE, 200);
	    if(ret == 0) break;
	    if(ret < 0) return 0;
	}

	free(write_buff);
	printf("OK\n");
    }

    write_buff = (__u8 *)malloc(BUFF_SIZE);
    memset(write_buff, 0, BUFF_SIZE);

    write_cmd[8] = SECT_NUM;

    for(;;) {

	bsize = read(in_f, write_buff, BUFF_SIZE);

	if(bsize < 0) {
	    perror("read");
	    free(write_buff);
	    exit(errno);
	}

	if((bsize == 0) || (bsize < BUFF_SIZE)) {
	    fprintf(stdout,"\rLBA: %6d\n", lba1); fflush(stdout);
	    free(write_buff);
	    return lba1;
	}

	fprintf(stdout,"\rLBA: %6d", lba1); fflush(stdout);
	lba = __swab32(lba1);
	memcpy((write_cmd + 2), (void *)&lba, 4);
	lba1 += SECT_NUM;

	while(1) {
	    ret = send_cmd(write_cmd, 10, SG_DXFER_TO_DEV, write_buff, BUFF_SIZE, 200);
	    if(ret == 0) break;
	    if(ret < 0) return 0;
	}
    }
}


int sync_cache()
{
    __u8 flush_cache[10];

    memset(flush_cache, 0, 10);
    flush_cache[0] = SYNCHRONIZE_CACHE;

    if(send_cmd(flush_cache, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int read_disk_info()
{
    __u8 read_disk_info_cmd[10];

    test_unit_ready();

    memset(DIB, 0, DIB_LEN);
    memset(read_disk_info_cmd, 0, 10);
    read_disk_info_cmd[0] = 0x51;
    read_disk_info_cmd[8] = DIB_LEN;

    if(send_cmd(read_disk_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)DIB, DIB_LEN, 20) < 0) {
	printf("\n    \n\n");
	return -1;
    }

    return 0;
}


int send_cue_sheet(char **argv, int trk_num)
{
    int i = 0, n;
    __u8 cue_sheet_cmd[10];
    __u8 *cue_buff = NULL;
    __u8 min, sec, frame;
    struct stat s;
    int cb_size = 0;
    int trk = -150;

    cb_size = trk_num * 16 + 16;

    cue_buff = (__u8 *)malloc(cb_size);
    if(cue_buff == NULL) {
	perror("cue_buff malloc");
	return -1;
    }

    memset(cue_sheet_cmd, 0, 10);
    memset(cue_buff, 0, cb_size);

/* Lead-In */
    cue_buff[0] = 0x01; cue_buff[1] = 0x00;
    cue_buff[2] = 0x00; cue_buff[3] = 0x01;
    cue_buff[4] = 0x00; cue_buff[5] = 0x00;
    cue_buff[6] = 0x00; cue_buff[7] = 0x00;

/* Pre-gap   */
    cue_buff[8] = 0x01; cue_buff[9] = 0x01;
    cue_buff[10] = 0x00; cue_buff[11] = 0x00;
    cue_buff[12] = 0x00; cue_buff[13] = 0x00;
    cue_buff[14] = 0x00; cue_buff[15] = 0x00;

/*   */
    cue_buff[16] = 0x01; cue_buff[17] = 0x01;
    cue_buff[18] = 0x01; cue_buff[19] = 0x00;
    cue_buff[20] = 0x00; cue_buff[21] = 0x00;
    cue_buff[22] = 0x02; cue_buff[23] = 0x00;


    for(i = 1, n = 24; i < trk_num; i++, n += 16) {

	memset((void *)&s, 0, sizeof(struct stat));
	if(stat(argv[i], &s)) {
    	    perror("stat");
	    return -1;
	}

	trk += s.st_size/CD_FRAMESIZE_RAW;
	min = LBA2MIN(trk);
	sec = LBA2SEC(trk, min);
	frame = LBA2FRAME(trk, min, sec);

/* Pre-Gap */
	cue_buff[n] = 0x01; cue_buff[n + 1] = i + 1;
	cue_buff[n + 2] = 0x00; cue_buff[n + 3] = 0x00;
	cue_buff[n + 4] = 0x00; cue_buff[n + 5] = min;
	cue_buff[n + 6] = sec; cue_buff[n + 7] = frame;

/* Track */
	sec += 2;
	cue_buff[n + 8] = 0x01; cue_buff[n + 9] = i + 1;
	cue_buff[n + 10] = 0x01; cue_buff[n + 11] = 0x00;
	cue_buff[n + 12] = 0x00; cue_buff[n + 13] = min;
	cue_buff[n + 14] = sec; cue_buff[n + 15] = frame;

#ifdef DEBUG
	printf("\nPre-Gap #%d MIN/SEC/FRAME: %.2d/%.2d/%.2d\n", i, min, sec - 2, frame);
	printf("Track #%d MIN/SEC/FRAME: %.2d/%.2d/%.2d\n", i, min, sec, frame);
#endif

    }

/* Lead-Out */
    memset((void *)&s, 0, sizeof(struct stat));
    if(stat(argv[trk_num], &s)) {
    	perror("stat");
	return;
    }

    trk += s.st_size/CD_FRAMESIZE_RAW + 150;
    min = LBA2MIN(trk);
    sec = LBA2SEC(trk, min);
    frame = LBA2FRAME(trk, min, sec);

#ifdef DEBUG
    printf("Lead-Out MIN/SEC/FRAME: %.2d/%.2d/%.2d\n", min, sec, frame);
#endif

    cue_buff[n] = 0x01; cue_buff[n + 1] = 0xAA;
    cue_buff[n + 2] = 0x01; cue_buff[n + 3] = 0x01;
    cue_buff[n + 4] = 0x00; cue_buff[n + 5] = min;
    cue_buff[n + 6] = sec; cue_buff[n + 7] = frame;

    cue_sheet_cmd[0] = 0x5D;
    cue_sheet_cmd[8] = cb_size;

    if(send_cmd(cue_sheet_cmd, 10, SG_DXFER_TO_DEV, cue_buff, cb_size, 200) < 0) return -1;

    free(cue_buff);
    return 0;
}


void do_record(int test_record_flag, char **argv, int trk_num)
{
    int i;
    time_t timv;
    __u32 track_size, block_num;
    int total_sect = 0;

#define LOCK 1
#define UNLOCK 0

    if(DIB->disk_status == 2) {
	printf("\n .   .\n\n");
	exit(0);
    }

    if(check_feature(CD_SAO) != 1) {
	printf("\n     SAO.\n\n");
	return;
    }

    if(mode_select(test_record_flag) < 0) {
	printf(" o  \n");
	return;
    }

    printf("Sending CUE SHEET..");
    if(send_cue_sheet(argv, trk_num) < 0) {
	printf("\n CUE SHEET\n");
	return;
    }
    printf("OK\n");

    block_num = MSF2LBA(DIB->l_out[1], DIB->l_out[2], DIB->l_out[3]);
    printf("\n    - %u\n", block_num);

    prevent_allow(LOCK);

    timv = time(0);
    printf("  : %s", ctime(&timv));

    for(i = 1; i <= trk_num; i++) {

	printf("\nWriting track #%d:\n", i);

	printf("Start sector - %u\n", total_sect);
	total_sect = write_audio(argv[i], i);

	if(total_sect == 0) {
	    printf("Cannot write %s\n", argv[i]);
	    return;
	}
    }

    if(sync_cache() < 0) printf("\n o \n");

    timv = time(0);
    printf("  : %s\n", ctime(&timv));

    prevent_allow(UNLOCK);
    eject_cd();

    return;
}


int main(int argc, char **argv)
{
    int i, type, test_record_flag = 0;
    char n;
    __u8 dev_name[9];

    memset(dev_name, 0, 9);
    memcpy(dev_name, SG_DEV, 8);

    printf("\n  :\n\n");

    for(i = 0; i < 4; i++) {

	if((sg_fd = open(dev_name, O_RDONLY)) < 0) continue;
	else {
	    printf("%d. ", i + 1);
	    if(inquiry() < 0) printf("\n   \n");
	}

	dev_name[7] += 1;

	close(sg_fd);
    }

    if(dev_name[7] == 0x30) {
	printf("    .\n\n");
	exit(0);
    }

    printf("\n : ");
    scanf("%c", &n);

    dev_name[7] = n - 1;
    if((sg_fd = open(dev_name, O_RDWR)) < 0) {
    	printf(" \n", dev_name);
	exit(-1);
    }


    DIB = (disk_info_block_t *)malloc(DIB_LEN);
    if(DIB == NULL) {
	perror("malloc");
	exit(errno);
    }

/*     */
    if(read_disk_info() < 0) goto out;

    printf("\n  :\n");
    printf("1.  .\n");
    printf("2.    .\n");
    printf("3.  .\n");
    printf(" : ");
    scanf("%d", &i);

    switch(i) {

	case 1:
	    blank();
	    break;

	case 2:
	    test_record_flag = TEST_RECORD_OFF; // 0

rec:
	    if(argc < 2) goto out;
	    printf("argc - %d\n", argc);
	    do_record(test_record_flag, argv, argc - 1);
	    break;

	case 3:
	    test_record_flag = TEST_RECORD_ON; // 1
	    goto rec;
	    break;

	default:
	    printf("\n   .\n\n");
	    break;

    }

out:
    free(DIB);
    close(sg_fd);

    return 0;
}
