#include <stdio.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <linux/cdrom.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include "page5.h"
#include "trk_info.h"
#include "pma.h"

#define SG_DEV "/dev/sg0"
#define MSF2LBA(Min, Sec, Frame) (((Min * 60 + Sec) * 75 + Frame) - 150)
#define WAV_HEADER_SIZE 44
#define SECT_NUM 10
#define BUFF_SIZE (CD_FRAMESIZE_RAW * SECT_NUM)

int sg_fd;
trk_info_t *trk_info; // track information
wpm_t *wpm; //      
__u16 page5_len;
__u8 *page5_data;
__u8 pma_data_buff[0xFFFF]; //      PMA

int mode_sense()
{
    __u8 mode_sense_cmd[10];

/*      */
    page5_data = (__u8 *)malloc(2);
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = 2;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, 2, 200) < 0) return -1;

    *((__u8 *)&page5_len) = page5_data[1];
    *((__u8 *)&page5_len + 1) = page5_data[0];
    page5_len += 2;

#ifdef DEBUG
    printf("    - %d\n", page5_len);
#endif

    free(page5_data);
    page5_data = (__u8 *)malloc(page5_len);

    memset(page5_data, 0, page5_len);
    wpm = (void *)(page5_data + 8); //       

    /* Prepare MODE_SENSE_10 command */
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_sense_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, page5_len, 200) < 0) return -1;

    return 0;
}


int mode_select(__u8 dbt, __u8 track_mode)
{
    __u8 mode_select_cmd[10];

    if(mode_sense() < 0) {
	printf("Mode sense error\n");
	exit(-1);
    }

    wpm->write_type = 1; // 1 - TAO
    wpm->dbt = dbt;
    wpm->track_mode = track_mode;
    wpm->s_format = 0;
    wpm->multises = 0;

    /* Prepare MODE_SELECT_10 command */
    memset(mode_select_cmd, 0, 10);
    mode_select_cmd[0] = MODE_SELECT_10;
    mode_select_cmd[1] = 0x10;
    mode_select_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_select_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_select_cmd, 10, SG_DXFER_TO_DEV, page5_data, page5_len, 200) < 0) return -1;

    free(page5_data);
    return 0;
}


int write_audio(__u8 *file_name, __u32 start_lba)
{
    int in_f;
    __u8 write_cmd[10];
    __u8 write_buff[BUFF_SIZE];
    __u32 lba = 0, lba1 = 0;
    ssize_t bsize = 0;

    in_f = open(file_name, O_RDONLY);
    if(in_f < 0) { perror("open (write_audio)"); exit(errno); }

    lseek(in_f, WAV_HEADER_SIZE, 0);
    memset(write_buff, 0, BUFF_SIZE);

    lba1 = start_lba;

    /* Prepare WRITE_10 command */
    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = SECT_NUM;

    test_unit_ready(sg_fd);

    for(;;) {

	bsize = read(in_f, write_buff, BUFF_SIZE);

	if(bsize < 0) { perror("read"); return -1; }

	if(bsize == 0) {
	    fprintf(stdout,"\rLBA: %6d (%6d)\n", lba1 - start_lba, lba1);
	    fflush(stdout);
	    return 0;
	}

	if(bsize < BUFF_SIZE) break;

	fprintf(stdout,"\rLBA: %6d (%6d)", lba1 - start_lba, lba1);
	fflush(stdout);
	lba = __swab32(lba1);
	memcpy((write_cmd + 2), (void *)&lba, 4);
	lba1 += SECT_NUM;

	if(send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, \
			BUFF_SIZE, 200) < 0) return -1;
    }

    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = bsize / CD_FRAMESIZE_RAW;

    fprintf(stdout,"\rLBA: %6d (%6d)", lba1 - start_lba, lba1);
    fflush(stdout);
    lba = __swab32(lba1);
    memcpy((write_cmd + 2), (void *)&lba, 4);

    if(send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, \
		    bsize, 200) < 0) return -1;

    fprintf(stdout,"\rLBA: %6d (%6d)\n", lba1 + write_cmd[8] - start_lba, lba1 + write_cmd[8]);
    fflush(stdout);

    return 0;
}


int sync_cache()
{
    __u8 flush_cache[10];

    test_unit_ready(sg_fd);

    memset(flush_cache, 0, 10);
    flush_cache[0] = SYNCHRONIZE_CACHE;

    if(send_cmd(sg_fd, flush_cache, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int close_track(int trk)
{
    __u8 close_trk_cmd[10];

    if(sync_cache() < 0) printf("Track #%d: cannot synchronize cache!\n", trk);

    memset(close_trk_cmd, 0, 10);
    close_trk_cmd[0] = 0x5B;
    close_trk_cmd[2] = 1;
    close_trk_cmd[5] = trk;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, close_trk_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int close_session()
{
    __u8 close_sess_cmd[10];

    memset(close_sess_cmd, 0, 10);
    close_sess_cmd[0] = 0x5B;
    close_sess_cmd[2] = 2;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, close_sess_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int read_pma()
{
    int i, k;
    __u8 read_pma_cmd[10];
    __u16 pma_data_len = 0; //    PMA
    __u32 lba;
    int pma_entries = 0; //   PMA
    pma_t *pma;

    memset(pma_data_buff, 0, 0xFFFF);

    memset(read_pma_cmd, 0, 10);
    read_pma_cmd[0] = 0x43; // READ_TOC/PMA/ATIP
    read_pma_cmd[2] = 3; // Format Field = 011b, PMA
    read_pma_cmd[7] = 0xFF;
    read_pma_cmd[8] = 0xFF;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, read_pma_cmd, 10, SG_DXFER_FROM_DEV, pma_data_buff, 0xFFFF, 200) < 0) return -1;

    *((__u8 *)&pma_data_len) = pma_data_buff[1];
    *((__u8 *)&pma_data_len + 1) = pma_data_buff[0];
    printf("PMA data length: %d\n", pma_data_len + 2);

    pma_entries = (pma_data_len - 2)/11;
    printf("PMA entries: %d\n", pma_entries);

    pma = (pma_t *)malloc(pma_data_len);
    memset((void *)pma, 0, pma_data_len);
    memcpy((void *)pma, pma_data_buff + 4, pma_data_len);

    printf("Entry\tADR\tCTRL\tPoint\tZero\tMin\tSec\tFrame\tPMin\tPsec\tPFrame\tLBA\n");

#define PMIN(i) (pma + i)->pmin
#define PSEC(i) (pma + i)->psec
#define PFRAME(i) (pma + i)->pframe
#define ADR(i) (pma + i)->adr

    for(i = 0; i < pma_entries; i++) {
	printf("%d\t", i);
	printf("%X\t", ADR(i));
	printf("%X\t", (pma + i)->ctrl);
	printf("%X\t", (pma + i)->point);
	printf("%d\t", (pma + i)->zero);
	printf("%d\t", (pma + i)->min);
	printf("%d\t", (pma + i)->sec);
	printf("%d\t", (pma + i)->frame);
	printf("%d\t", PMIN(i));
	printf("%d\t", PSEC(i));
	printf("%d\t", PFRAME(i));

	lba = MSF2LBA(PMIN(i), PSEC(i), PFRAME(i));

	if(ADR(i) != 1) printf("---\n");
	else printf("%u\n", lba);
    }

    free(pma);
    return 0;
}


int reserv_track(__u32 track_size)
{
    __u8 reserv_track_cmd[10];
    __u32 size = 0;

    memset(reserv_track_cmd, 0, 10);
    reserv_track_cmd[0] = 0x53;

    size = __swab32(track_size);
    memcpy((void *)(reserv_track_cmd + 5), (void *)&size, 4);

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, reserv_track_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    read_pma();

    return 0;
}


int read_track_info(int trk_num)
{
    __u8 read_track_info_cmd[10];

    memset((void *)trk_info, 0, TRK_INFO_LEN);
    memset(read_track_info_cmd, 0, 10);
    read_track_info_cmd[0] = 0x52;
    read_track_info_cmd[1] = 1;
    read_track_info_cmd[5] = trk_num; // track number
    read_track_info_cmd[8] = TRK_INFO_LEN;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, read_track_info_cmd, 10, SG_DXFER_FROM_DEV, (void *)trk_info, TRK_INFO_LEN, 200) < 0) {
	printf("Cannot read track info!\n");
	return -1;
    }

    return 0;
}


/*    MP3  WAV */
void mp3_decoder(__u8 *file_name)
{
    static pid_t pid;
    int status;

    switch(pid = fork()) {

	case -1:
	    perror("fork");
	    exit(-1);

	case 0:
	    execl("/usr/bin/mpg321", "mpg321", "-q", "-w", "track.wav", file_name, 0);
	    exit(-1);
    }

    if((pid = waitpid(pid, &status, 0)) && WIFEXITED(status)) return;
}


void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);

    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    send_cmd(sg_fd, start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 200);

    return;
}


int main(int argc, char **argv)
{
    int i = 1;
    struct stat s;
    struct dirent *d;
    DIR *dp;
    __u8 full_path[255];
    __u32 start_lba = 0, track_size = 0;

    if(argc != 2) {
	printf("\n\tUsage: mp2cdda [Dir. with MP3-files]\n\n");
	return 0;
    }

    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return -1;
    }

/* dbt = 0, track_mode = 0 */
    if(mode_select(0, 0) < 0) {
	printf("Mode select error\n");
	return 0;
    }

    trk_info = (trk_info_t *)malloc(TRK_INFO_LEN);
    if(trk_info == NULL) {
	perror("malloc");
	exit(-1);
    }

    if((dp = opendir(argv[1])) == NULL) {
	perror("opendir");
	return -1;
    }

/*      */
    d = readdir(dp); // "."
    d = readdir(dp); // ".."

    for(i = 1;; i++) {

	if((d = readdir(dp)) == NULL) break;
	printf("\n%d. %s\n", i, d->d_name);

	memset(full_path, 0, 255);
	sprintf(full_path, "%s/%s", argv[1], d->d_name);

	printf("\nDecoding file %s..\n", d->d_name);
	mp3_decoder(full_path);
	printf("OK\n");

	memset((void *)&s, 0, sizeof(struct stat));
	if(stat("./track.wav", &s)) {
	    perror("stat");
	    return -1;
	}

	track_size = (s.st_size - 44)/ CD_FRAMESIZE_RAW;

    /*     */
	if(reserv_track(track_size) < 0) {
	    printf("Cannot reserve track %d\n", i);
	    continue;
	}

	if(read_track_info(i) < 0) return -1;
	start_lba = __swab32(trk_info->start_addr);

	printf("Track number: %d\n", i);
	printf("Track size: %u\n", track_size);
	printf("Start LBA of track #%d: %u\n", i, start_lba);

	if(write_audio("./track.wav", start_lba) < 0) {
	    printf("Cannot write %s\n", d->d_name);
	    return -1;
	}

	if(close_track(i) < 0) {
	    printf("Cannot close track #%d!\n", i);
    	    return -1;
	}
    }

    fprintf(stdout, "Close session..\n");

    if(close_session() < 0) printf("\nCannot close session!\n");
    else printf("OK\n\n");

    closedir(dp);
    unlink("./track.wav");
    free(trk_info);
    eject_cd();

    close(sg_fd);

    return 0;
}
