#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <linux/cdrom.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <sys/stat.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include "iso2cd_multi.h"

#define SG_DEV "/dev/sg1"

int sg_fd;
__u16 page5_len;
__u8 *page5_data;
wpm_t *wpm; //    
disk_info_block_t *DIB;


int mode_sense()
{
    __u8 mode_sense_cmd[10];

/*      */
    page5_data = (__u8 *)malloc(2);
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = 2;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, 2, 200) < 0) return -1;

    *((__u8 *)&page5_len) = page5_data[1];
    *((__u8 *)&page5_len + 1) = page5_data[0];
    page5_len += 2;

#ifdef DEBUG
    printf("    - %d\n", page5_len);
#endif

    free(page5_data);
    page5_data = (__u8 *)malloc(page5_len);

    memset(page5_data, 0, page5_len);
    wpm = (void *)(page5_data + 8); //       

    /* Prepare MODE_SENSE_10 command */
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = page5_len;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, page5_len, 200) < 0) return -1;

    return 0;
}


int mode_select(__u8 dbt, __u8 track_mode)
{
    __u8 mode_select_cmd[10];

    if(mode_sense() < 0) {
	printf("Mode sense error\n");
	exit(-1);
    }

    wpm->write_type = 1; // 1- TAO, 2 - SAO (DAO for DVD)
    wpm->dbt = dbt; // 8
    wpm->s_format = 0;
    wpm->track_mode = track_mode; // 4
    wpm->multises = 3; //    

    /* Prepare MODE_SELECT_10 command */
    memset(mode_select_cmd, 0, 10);
    mode_select_cmd[0] = MODE_SELECT_10;
    mode_select_cmd[1] = 0x10;
    mode_select_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_select_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_select_cmd, 10, SG_DXFER_TO_DEV, page5_data, page5_len, 200) < 0) return -1;

    free(page5_data);
    return 0;
}


/*     */
int inquiry()
{

#define INQ_SIZE 40

    __u8 inq_cmd[6];
    __u8 data_buff[INQ_SIZE];

    memset(inq_cmd, 0, 6);
    memset(data_buff, 0, INQ_SIZE);
    inq_cmd[0] = INQUIRY;
    inq_cmd[4] = INQ_SIZE; //   

    if(send_cmd(sg_fd, inq_cmd, 6, SG_DXFER_FROM_DEV, data_buff, INQ_SIZE, 200) < 0) return -1;

    printf("\n%.8s%.16s%.4s\n", data_buff + 8, data_buff + 16, data_buff + 32);

    return 0;
}


/*    CD Track at Once */
int check_feature(__u16 f_num)
{

#define FEATURE_LEN 16

    int i;
    __u8 get_conf_cmd[10]; // CDB
    __u8 data_buff[FEATURE_LEN]; //  
    __u32 data_length = 0; //   

#ifdef DEBUG
    __u16 cur_prof = 0; //  , Current Profile
#endif

    cd_tao_t *cd_tao;

    memset(data_buff, 0, sizeof(data_buff));
    cd_tao = (void *)(data_buff + 8);

    memset(get_conf_cmd, 0, 10);
    get_conf_cmd[0] = 0x46; // GET CONFIGURATION
    get_conf_cmd[1] = 2; // RT = 10b
    get_conf_cmd[2] = *((__u8 *)&f_num + 1);
    get_conf_cmd[3] = *((__u8 *)&f_num);
    get_conf_cmd[8] = FEATURE_LEN;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, get_conf_cmd, 10, SG_DXFER_FROM_DEV, data_buff, FEATURE_LEN, 200) < 0) {
	printf("   \n\n");
	return -1;
    }

    memcpy((void *)&data_length, data_buff, 4);
    data_length = __swab32(data_length);
    if(data_length == 4) return -1; //   

#ifdef DEBUG
/*   */
    *((__u8 *)&cur_prof) = data_buff[7];
    *((__u8 *)&cur_prof + 1) = data_buff[6];
    printf("\n : 0x%04X\n", cur_prof);
#endif

    return (cd_tao->cd_rw);
}


/* /  */
int prevent_allow(__u8 prevent_flag)
{
    __u8 prevent_allow_cmd[6];

    memset(prevent_allow_cmd, 0, 6);
    prevent_allow_cmd[0] = 0x1E;
    prevent_allow_cmd[4] = prevent_flag;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, prevent_allow_cmd, 6, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int write_iso(__u8 *file_name, __u32 start_lba)
{
    int ret, in_f;
    __u8 write_cmd[10];
    __u8 write_buff[BUFF_SIZE];
    __u32 lba = 0, lba1 = 0;
    ssize_t bsize = 0;

    in_f = open(file_name, O_RDONLY);
    if(in_f < 0) { perror("open"); exit(errno); }

    memset(write_buff, 0, BUFF_SIZE);

    lba1 = start_lba;

    /* Prepare WRITE_10 command */
    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = SECT_NUM;

    test_unit_ready(sg_fd);

    for(;;) {

	bsize = read(in_f, write_buff, BUFF_SIZE);

	if(bsize < 0) { perror("read"); exit(errno); }

	if(bsize == 0) {
	    fprintf(stdout,"\rLBA: %6d (%6d)\n", lba1 - start_lba, lba1);
	    fflush(stdout);
	    return 0;
	}

	if(bsize < BUFF_SIZE) break;

	fprintf(stdout,"\rLBA: %6d (%6d)", lba1 - start_lba, lba1);
	fflush(stdout);
	lba = __swab32(lba1);
	memcpy((write_cmd + 2), (void *)&lba, 4);
	lba1 += SECT_NUM;

	ret = send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, BUFF_SIZE, 200);
	if(ret < 0) return -1;
    }

    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = bsize / CD_FRAMESIZE;

    fprintf(stdout,"\rLBA: %6d (%6d)", lba1 - start_lba, lba1);
    fflush(stdout);
    lba = __swab32(lba1);
    memcpy((write_cmd + 2), (void *)&lba, 4);

    ret = send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, bsize, 200);
    if(ret < 0) return -1;

    fprintf(stdout,"\rLBA: %6d (%6d)\n", lba1 + write_cmd[8] - start_lba, lba1 + write_cmd[8]);
    fflush(stdout);

    return 0;
}


int sync_cache()
{
    __u8 flush_cache[10];

    memset(flush_cache, 0, 10);
    flush_cache[0] = SYNCHRONIZE_CACHE;

    if(send_cmd(sg_fd, flush_cache, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int close_track(int trk_num)
{
    __u8 close_trk_cmd[10];

    if(sync_cache() < 0) printf("\n-ERR: SYNCHRONIZE CACHE\n");

    memset(close_trk_cmd, 0, 10);
    close_trk_cmd[0] = 0x5B;
    close_trk_cmd[2] = 1;
    close_trk_cmd[5] = trk_num; //  

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, close_trk_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int close_session()
{
    __u8 close_sess_cmd[10];

    memset(close_sess_cmd, 0, 10);
    close_sess_cmd[0] = 0x5B;
    close_sess_cmd[2] = 2;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, close_sess_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


int read_disk_info()
{
    __u8 read_disk_info_cmd[10];

    memset(DIB, 0, DIB_LEN);
    memset(read_disk_info_cmd, 0, 10);
    read_disk_info_cmd[0] = 0x51;
    read_disk_info_cmd[8] = DIB_LEN;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, read_disk_info_cmd, 10, SG_DXFER_FROM_DEV, (__u8 *)DIB, DIB_LEN, 200) < 0) {
	printf("\n    \n\n");
	return -1;
    }

    return 0;
}


int reserv_track(__u32 track_size)
{
    __u8 reserv_track_cmd[10];
    __u32 size = 0;

    memset(reserv_track_cmd, 0, 10);
    reserv_track_cmd[0] = 0x53;

    size = __swab32(track_size);
    memcpy((void *)(reserv_track_cmd + 5), (void *)&size, 4);

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, reserv_track_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


__u32 read_track_info(int trk_num)
{
    __u8 read_track_info_cmd[10];
    __u8 data_buff[40];
    __u32 lba = 0;

    memset(data_buff, 0, 40);
    memset(read_track_info_cmd, 0, 10);
    read_track_info_cmd[0] = 0x52;
    read_track_info_cmd[1] = 1;
    read_track_info_cmd[5] = trk_num;
    read_track_info_cmd[7] = 40;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, read_track_info_cmd, 10, SG_DXFER_FROM_DEV, data_buff, 40, 200) < 0) {
	printf("\n     #%d\n\n", trk_num);
	exit(-1);
    }

    memcpy((void *)&lba, (void *)(data_buff + 8), 4);

    return __swab32(lba);
}


/*   */
void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);

    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    send_cmd(sg_fd, start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 200);

    return;
}



int read_raw_toc(int flag)
{

/* :
 * flag == 0 -      
 * flag == 1 -  RAW TOC
 */

    int i;
    __u8 read_toc_cmd[10];
    __u8 *data_buff; //      TOC
    __u16 toc_data_len = 0; //   TOC
    __u32 lba;
    int toc_entries = 0; //   TOC
    int first_trk = 0; //     
    int last = 0; //   /
    toc_t *t;

    data_buff = (__u8 *)malloc(0xFFFF);
    if(data_buff == NULL) { perror("malloc"); exit(errno); }

    memset(data_buff, 0, 0xFFFF);
    memset(read_toc_cmd, 0, 10);

    read_toc_cmd[0] = READ_TOC;
    if(flag) read_toc_cmd[2] = 2; // Format Field = 10b - RAW TOC
    read_toc_cmd[7] = 0xFF;
    read_toc_cmd[8] = 0xFF;

#ifdef DEBUG
    printf("Format Field = %d\n", read_toc_cmd[2]);
#endif

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, read_toc_cmd, 10, SG_DXFER_FROM_DEV, data_buff, 0xFFFF, 200) < 0) return -1;

    last = data_buff[3];
    if(!flag) { free(data_buff); return last; }

    *((__u8 *)&toc_data_len) = data_buff[1];
    *((__u8 *)&toc_data_len + 1) = data_buff[0];

    toc_entries = (toc_data_len - 2)/11;

    t = (toc_t *)malloc(toc_data_len);
    memset((void *)t, 0, toc_data_len);
    memcpy((void *)t, data_buff + 4, toc_data_len);

    free(data_buff);

#define SNUM(i) (t + i)->snum
#define PMIN(i) (t + i)->pmin
#define PSEC(i) (t + i)->psec
#define PFRAME(i) (t + i)->pframe
#define MIN(i) (t + i)->min
#define SEC(i) (t + i)->sec
#define FRAME(i) (t + i)->frame
#define POINT(i) (t + i)->point

    printf("\n   : %d\n", last);
    printf("Entry\tSession\tPoint\tMin\tSec\tFrame\tPMin\tPsec\tPFrame\tLBA\n");

    for(i = 0; i < toc_entries; i++) {
	printf("%d\t", i);
	printf("%d\t", SNUM(i)); //(t + i)->snum);
	printf("%X\t", POINT(i)); //(t + i)->point);
	printf("%d\t", MIN(i)); //(t + i)->min);
	printf("%d\t", SEC(i)); //(t + i)->sec);
	printf("%d\t", FRAME(i)); //(t + i)->frame);
	printf("%d\t", PMIN(i)); //(t + i)->pmin);
	printf("%d\t", PSEC(i)); //(t + i)->psec);
	printf("%d\t", PFRAME(i)); //(t + i)->pframe);

	if((POINT(i) == 0xC0) || (POINT(i) == 0xC1)) {
	    printf("\n");
	    continue;
	}

	if(POINT(i) == 0xB0) lba = MSF2LBA(MIN(i), SEC(i), FRAME(i)) + 150;
	else lba = MSF2LBA(PMIN(i), PSEC(i), PFRAME(i));

	printf("%u\n", lba);
    }

    for(i = 0; i < toc_entries; i++) {
	if((POINT(i) == 0xA0) && (SNUM(i) == last)) {
	    first_trk = PMIN(i); //     
	    break;
	}
    }

    for(i = 0; i < toc_entries; i++) {
	if((POINT(i) == first_trk) && (SNUM(i) == last)) {
	    lba = MSF2LBA(PMIN(i), PSEC(i), PFRAME(i));
	    printf("\n%u\t", lba);
	    break;
	}
    }

    for(i = 0; i < toc_entries; i++) {
	if((POINT(i) == 0xB0) && (SNUM(i) == last)){
	    lba = MSF2LBA(MIN(i), SEC(i), FRAME(i)) + 150;
	    printf("%u\n\n", lba);
	    break;
	}
    }

    free(t);
    return 0;
}



int read_ms_info()
{
    __u8 read_toc_cmd[10];
    __u8 data_buff[12];
    ms_info_t *ms_info;

    memset(data_buff, 0, 12);
    ms_info = (void *)(data_buff + 4);

    memset(read_toc_cmd, 0, 10);    
    read_toc_cmd[0] = READ_TOC;
    read_toc_cmd[2] = 1;
    read_toc_cmd[8] = 12;

    if(send_cmd(sg_fd, read_toc_cmd, 10, SG_DXFER_FROM_DEV, data_buff, 12, 200) < 0) {
	printf("\n    \n\n");
	return -1;
    }

    printf("  :\n");
    printf("%u\t", __swab32(ms_info->start_addr));
    return 0;
}



int main(int argc, char **argv)
{
    int i = 1;
    int trk_num = 1; //   
    struct stat s;
    time_t timv;
    __u32 track_size, block_num, start_lba;

#define LOCK 1
#define UNLOCK 0

    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
        perror("open");
        return 1;
    }

    if(inquiry() < 0) {
	printf("\n    \n");
	exit(-1);
    }

    if(check_feature(CD_TAO) != 1) {
	printf("\n     CD-R/RW .\n\n");
	return 0;
    }

    DIB = (disk_info_block_t *)malloc(DIB_LEN);
    if(DIB == NULL) {
	perror("malloc");
	exit(errno);
    }

/*     */
    if(read_disk_info() < 0) exit(-1);

    if(DIB->disk_status == 2) { //   .  .
	printf("\n .  .\n\n");
	goto out;
    }

    if(argc == 1) {

	if(DIB->disk_status == 0) printf("\n \n\n");
	else {
	    if(read_raw_toc(1) < 0) printf("\n  TOC\n\n");
	    read_ms_info();
	    start_lba = read_track_info(0xFF);
	    printf("%u\n\n", start_lba);
	}

	close(sg_fd);
	return 0;
    }

    if(DIB->disk_status == 1) {

	trk_num = read_raw_toc(0); //   

	if(trk_num < 0) {
	    printf("\n  TOC\n\n");
	    close(sg_fd);
	    return -1;
	}

	trk_num += 1;
    }

    block_num = MSF2LBA(DIB->l_out[1], DIB->l_out[2], DIB->l_out[3]);

    memset((void *)&s, 0, sizeof(struct stat));
    if(stat(argv[1], &s)) {
        perror("stat");
        return -1;
    }

    track_size = s.st_size/CD_FRAMESIZE;

    printf("\n   : %u\n", block_num);
    printf("  %s: %u K ", argv[1], s.st_size/1024);
    printf("(%u )\n", track_size);

    if(track_size > block_num) {
	printf("\n - %s    \n\n", argv[1]);
	exit(0);
    }

    if(mode_select(8, 4) < 0) {
	printf(" o  \n");
	return 0;
    }

    if(reserv_track(track_size) < 0) {
        printf("\n  \n");
        return -1;
    }

    start_lba = read_track_info(trk_num);

    timv = time(0);
    printf("  : %s", ctime(&timv));

    prevent_allow(LOCK);

    if(write_iso(argv[1], start_lba) < 0) {
	printf("  - %s\n", argv[1]);
	goto out1;
    }

    if(close_track(trk_num) < 0) {
	printf("  \n");
	goto out1;
    }

    fprintf(stdout,"\n ..");
    fflush(stdout);
    if(close_session() < 0) printf("  \n");
    else {
	fprintf(stdout,"OK\n");
	timv = time(0);
	fprintf(stdout,"  : %s\n", ctime(&timv));
	fflush(stdout);
    }

out1:
    prevent_allow(UNLOCK);
    eject_cd();

out:
    free(DIB);
    close(sg_fd);

    return 0;
}
