#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <linux/cdrom.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <sys/stat.h>
#include <linux/types.h>
#include <linux/byteorder/swab.h>
#include "iso2cd_tao.h"

#define SG_DEV "/dev/sg0"

int sg_fd;
__u16 page5_len;
__u8 *page5_data;
wpm_t *wpm; //    

int mode_sense()
{
    __u8 mode_sense_cmd[10];

/*      */
    page5_data = (__u8 *)malloc(2);
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = 2;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, 2, 200) < 0) return -1;

    *((__u8 *)&page5_len) = page5_data[1];
    *((__u8 *)&page5_len + 1) = page5_data[0];
    page5_len += 2;

#ifdef DEBUG
    printf("    - %d \n", page5_len);
#endif

    free(page5_data);
    page5_data = (__u8 *)malloc(page5_len);

    memset(page5_data, 0, page5_len);
    wpm = (void *)(page5_data + 8); //       

    /* Prepare MODE_SENSE_10 command */
    memset(mode_sense_cmd, 0, 10);
    mode_sense_cmd[0] = MODE_SENSE_10;
    mode_sense_cmd[2] = 5; //   
    mode_sense_cmd[8] = page5_len;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_sense_cmd, 10, SG_DXFER_FROM_DEV, page5_data, page5_len, 200) < 0) return -1;

    return 0;
}


int mode_select(__u8 dbt, __u8 track_mode)
{
    __u8 mode_select_cmd[10];

    if(mode_sense() < 0) {
	printf("Mode sense error\n");
	exit(-1);
    }

    wpm->write_type = 1; // 1- TAO, 2 - SAO (DAO for DVD)
    wpm->dbt = dbt; // 8
    wpm->s_format = 0;
    wpm->track_mode = track_mode; // 4

    /* Prepare MODE_SELECT_10 command */
    memset(mode_select_cmd, 0, 10);
    mode_select_cmd[0] = MODE_SELECT_10;
    mode_select_cmd[1] = 0x10;
    mode_select_cmd[7] = *((__u8 *)&page5_len + 1);
    mode_select_cmd[8] = *(__u8 *)&page5_len;

    test_unit_ready(sg_fd);

    if(send_cmd(sg_fd, mode_select_cmd, 10, SG_DXFER_TO_DEV, page5_data, page5_len, 200) < 0) return -1;

    free(page5_data);
    return 0;
}


/*     */
int inquiry()
{
    __u8 inq_cmd[6];
    __u8 data_buff[40];

    memset(inq_cmd, 0, 6);
    memset(data_buff, 0, 40);
    inq_cmd[0] = INQUIRY;
    inq_cmd[4] = 40; //   

    if(send_cmd(sg_fd, inq_cmd, 6, SG_DXFER_FROM_DEV, data_buff, 40, 200) < 0) return -1;

    printf("\n%.8s%.16s%.4s\n", data_buff + 8, data_buff + 16, data_buff + 32);

    return 0;
}


/* /  */
int prevent_allow(__u8 prevent_flag)
{
    __u8 prevent_allow_cmd[6];

    memset(prevent_allow_cmd, 0, 6);
    prevent_allow_cmd[0] = 0x1E;
    prevent_allow_cmd[4] = prevent_flag;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, prevent_allow_cmd, 6, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


/*      */
void eject_cd()
{
    __u8 start_stop_cmd[6];

    memset(start_stop_cmd, 0, 6);

    start_stop_cmd[0] = 0x1B;
    start_stop_cmd[4] = 2;

    send_cmd(sg_fd, start_stop_cmd, 6, SG_DXFER_NONE, NULL, 0, 200);

    return;
}


/*    */
int write_iso(__u8 *file_name)
{
    int ret, in_f;
    __u8 write_cmd[10];
    __u8 write_buff[BUFF_SIZE];
    __u32 lba = 0, lba1 = 0;
    ssize_t bsize = 0;

    test_unit_ready(sg_fd);

    in_f = open(file_name, O_RDONLY);
    if(in_f < 0) { perror("open"); exit(errno); }

    memset(write_buff, 0, BUFF_SIZE);

    /* Prepare WRITE_10 command */
    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = SECT_NUM;

    for(;;) {

	bsize = read(in_f, write_buff, BUFF_SIZE);

	if(bsize < 0) { perror("read"); exit(errno); }

	if(bsize == 0) {
	    fprintf(stdout,"\rLBA: %6d\n", lba1);
	    fflush(stdout);
	    return 0;
	}

	if(bsize < BUFF_SIZE) break;

	fprintf(stdout,"\rLBA: %6d", lba1); fflush(stdout);
	lba = __swab32(lba1);
	memcpy((write_cmd + 2), (void *)&lba, 4);
	lba1 += SECT_NUM;

	ret = send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, BUFF_SIZE, 200);
	if(ret < 0) return -1;
    }

    memset(write_cmd, 0, 10);
    write_cmd[0] = WRITE_10;
    write_cmd[8] = bsize / CD_FRAMESIZE;

    fprintf(stdout,"\rLBA: %6d", lba1); fflush(stdout);
    lba = __swab32(lba1);
    memcpy((write_cmd + 2), (void *)&lba, 4);

    ret = send_cmd(sg_fd, write_cmd, 10, SG_DXFER_TO_DEV, write_buff, bsize, 200);
    if(ret < 0) return -1;

    fprintf(stdout,"\rLBA: %6d\n", lba1 + write_cmd[8]);
    fflush(stdout);

    return 0;
}


int sync_cache()
{
    __u8 flush_cache[10];

    memset(flush_cache, 0, 10);
    flush_cache[0] = SYNCHRONIZE_CACHE;

    if(send_cmd(sg_fd, flush_cache, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


/*   */
int close_track()
{
    __u8 close_trk_cmd[10];

    if(sync_cache() < 0) printf("\n-ERR: SYNCHRONIZE CACHE\n");

    memset(close_trk_cmd, 0, 10);
    close_trk_cmd[0] = 0x5B;
    close_trk_cmd[2] = 1;
    close_trk_cmd[5] = 1; //  

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, close_trk_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


/*   */
int close_session()
{
    __u8 close_sess_cmd[10];

    memset(close_sess_cmd, 0, 10);
    close_sess_cmd[0] = 0x5B;
    close_sess_cmd[2] = 2;

    test_unit_ready(sg_fd);
    if(send_cmd(sg_fd, close_sess_cmd, 10, SG_DXFER_NONE, NULL, 0, 60) < 0) return -1;

    return 0;
}


/*   */
int reserv_track(__u32 track_size)
{
    __u8 reserv_track_cmd[10];
    __u32 size = 0;

    test_unit_ready(sg_fd);

    memset(reserv_track_cmd, 0, 10);
    reserv_track_cmd[0] = 0x53;

    size = __swab32(track_size);
    memcpy((void *)(reserv_track_cmd + 5), (void *)&size, 4);

    if(send_cmd(sg_fd, reserv_track_cmd, 10, SG_DXFER_NONE, NULL, 0, 200) < 0) return -1;

    return 0;
}


void do_record(__u8 *image_name)
{
    struct stat s;
    time_t timv;
    __u32 track_size;

    memset((void *)&s, 0, sizeof(struct stat));
    if(stat(image_name, &s)) { perror("stat"); return; }

    track_size = s.st_size/CD_FRAMESIZE;

    printf("  %s: %u K", image_name, s.st_size/1024);
    printf(" (%u )\n", track_size);

    if(mode_select(8, 4) < 0) {
	printf(" o  \n");
	return;
    }

    if(reserv_track(track_size) < 0) {
        printf("\n  \n");
        return;
    }

    timv = time(0);
    printf("  : %s", ctime(&timv));

    prevent_allow(LOCK);

    if(write_iso(image_name) < 0) {
	printf("  - %s\n", image_name);
	return;
    }

    if(close_track() < 0) {
	printf("  \n");
	return;
    }

    printf("\n ..\n");
    if(close_session() < 0) {
	printf("  \n");
	return;
    }
    printf("OK\n");

    timv = time(0);
    printf("  : %s\n", ctime(&timv));

    prevent_allow(UNLOCK);

    eject_cd();

    return;
}


int main(int argc, char **argv)
{
    if((sg_fd = open(SG_DEV, O_RDWR)) < 0) {
	perror("open");
	return -1;
    }

    if(inquiry() < 0) {
	printf("\n   \n");
	return -1;
    }

    if(argc != 2) {
    	printf("\n     -:");
    	printf(" iso2cd [ISO-]\n\n");
	return -1;
    }

    do_record(argv[1]);

    close(sg_fd);

    return 0;
}
